from typing import Literal
from pyteal import *
from beaker import *


HashValue = abi.StaticBytes[Literal[32]]
Signature = abi.StaticBytes[Literal[65]]


class EthEcdsaVerify(LogicSignature):
    """
    This Lsig has a single method  `eth_ecdsa_validate` that takes two application arguments:
      hash, signature
    and returns the validity of the signature given the hash
    as written in OpenZeppelin https://docs.openzeppelin.com/contracts/2.x/api/cryptography#ECDSA-recover-bytes32-bytes-
    (65-byte signatures only)
    """

    def evaluate(self):
        return Seq(
            Assert(
                # Don't let it be rekeyed
                Txn.rekey_to() == Global.zero_address(),
                # Don't take a fee from me
                Txn.fee() == Int(0),
                # Make sure I've signed an app call
                Txn.type_enum() == TxnType.ApplicationCall,
                # Make sure I have the args I expect [method_selector, hash_value, signature]
                Txn.application_args.length() == Int(3),
            ),
            self.eth_ecdsa_validate(Txn.application_args[1], Txn.application_args[2]),
        )

    @internal(TealType.uint64)
    def eth_ecdsa_validate(self, hash_value: Expr, signature: Expr) -> Expr:
        """
        Return a 1/0 for valid signature given hash

        Equivalent of OpenZeppelin ECDSA.recover for long 65-byte Ethereum signatures
        https://docs.openzeppelin.com/contracts/2.x/api/cryptography#ECDSA-recover-bytes32-bytes-
        Short 64-byte Ethereum signatures require some changes to the code


        [1] https://github.com/OpenZeppelin/openzeppelin-contracts/blob/5fbf494511fd522b931f7f92e2df87d671ea8b0b/contracts/utils/cryptography/ECDSA.sol#L153


        Note: Unless compatibility with Ethereum or another system is necessary,
        we highly recommend using ed25519_verify instead of ecdsa on Algorand

        WARNING: This code has NOT been audited
        DO NOT USE IN PRODUCTION
        """

        r = Extract(signature, Int(0), Int(32))
        s = Extract(signature, Int(32), Int(32))

        # The recovery ID is shifted by 27 on Ethereum
        # For non-Ethereum signatures, remove the -27 on the line below
        v = Btoi(Extract(signature, Int(64), Int(1))) - Int(27)

        return Seq(
            Assert(
                Len(signature) == Int(65),
                Len(hash_value) == Int(32),
                # The following two asserts are to prevent malleability like in [1]
                BytesLe(
                    s,
                    Bytes(
                        "base16",
                        "0x7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF5D576E7357A4501DDFE92F46681B20A0",
                    ),
                ),
                v <= Int(1),
            ),
            EcdsaVerify(
                EcdsaCurve.Secp256k1,
                hash_value,
                r,
                s,
                EcdsaRecover(EcdsaCurve.Secp256k1, hash_value, v, r, s),
            ),
        )



import pytest
import pyteal as pt

from .decorators import (
    external,
    get_handler_config,
    DefaultArgument,
    Authorize,
    create,
    clear_state,
    close_out,
    delete,
    update,
    no_op,
    opt_in,
)

options = pt.CompileOptions(mode=pt.Mode.Application, version=pt.MAX_TEAL_VERSION)


def test_handler_config():
    @external
    def handleable():
        pass

    hc = get_handler_config(handleable)

    assert hc.method_spec is not None, "Expected abi method to be created"
    meth = hc.method_spec
    assert len(meth.args) == 0, "Expected no args"
    assert meth.name == "handleable", "Expected name to match"

    config_dict = hc.__dict__
    del config_dict["method_spec"]

    for k, v in config_dict.items():
        assert v is None or v is False, f"Expected {k} to be unset"

    ###

    @external(read_only=True)
    def handleable():
        pass

    hc = get_handler_config(handleable)

    config_dict = hc.__dict__

    assert hc.method_spec is not None, "Expected abi method to be created"
    del config_dict["method_spec"]

    assert hc.read_only is True, "Expected read_only to be true"
    del config_dict["read_only"]

    for k, v in config_dict.items():
        assert v is None or v is False, f"Expected {k} to be unset"

    ###

    @external(authorize=Authorize.only(pt.Global.creator_address()))
    def handleable():
        pass

    hc = get_handler_config(handleable)

    config_dict = hc.__dict__

    assert hc.method_spec is not None, "Expected abi method to be created"
    del config_dict["method_spec"]
    for k, v in config_dict.items():
        assert v is None or v is False, f"Expected {k} to be unset"

    ###

    @external(method_config=pt.MethodConfig(opt_in=pt.CallConfig.CALL))
    def handleable():
        pass

    hc = get_handler_config(handleable)

    config_dict = hc.__dict__

    assert hc.method_spec is not None, "Expected abi method to be created"
    del config_dict["method_spec"]

    assert hc.method_config is not None, "Expected method config to be set"
    assert (
        hc.method_config.opt_in == pt.CallConfig.CALL
    ), "Expected method config opt in to be set to call"
    del config_dict["method_config"]

    for k, v in config_dict.items():
        assert v is None or v is False, f"Expected {k} to be unset"


def test_authorize():

    cmt = "unauthorized"

    auth_only = Authorize.only(pt.Global.creator_address())

    expr = pt.Txn.sender() == pt.Global.creator_address()
    expected = expr.__teal__(options)
    actual = auth_only.subroutine.implementation(pt.Txn.sender()).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    @external(authorize=auth_only)
    def creator_only():
        return pt.Approve()

    expr = pt.Seq(pt.Assert(auth_only(pt.Txn.sender()), comment=cmt), pt.Approve())

    expected = expr.__teal__(options)
    actual = creator_only().__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealTypeError):
        Authorize.only(pt.Int(1))

    asset_id = pt.Int(123)
    auth_holds_token = Authorize.holds_token(asset_id)

    balance = pt.AssetHolding.balance(pt.Txn.sender(), asset_id)
    expr = pt.Seq(balance, pt.And(balance.hasValue(), balance.value() > pt.Int(0)))
    expected = expr.__teal__(options)
    actual = auth_holds_token.subroutine.implementation(pt.Txn.sender()).__teal__(
        options
    )

    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    @external(authorize=auth_holds_token)
    def holds_token_only():
        return pt.Approve()

    expr = pt.Seq(
        pt.Assert(auth_holds_token(pt.Txn.sender()), comment=cmt), pt.Approve()
    )

    expected = expr.__teal__(options)
    actual = holds_token_only().__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealTypeError):
        Authorize.holds_token(pt.Bytes("abc"))

    app_id = pt.Int(123)
    auth_opted_in = Authorize.opted_in(app_id)

    expr = pt.App.optedIn(pt.Txn.sender(), app_id)

    expected = expr.__teal__(options)
    actual = auth_opted_in.subroutine.implementation(pt.Txn.sender()).__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    @external(authorize=auth_opted_in)
    def opted_in_only():
        return pt.Approve()

    expr = pt.Seq(pt.Assert(auth_opted_in(pt.Txn.sender()), comment=cmt), pt.Approve())

    expected = expr.__teal__(options)
    actual = opted_in_only().__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    # Bare handler

    @delete(authorize=auth_only)
    def deleter():
        return pt.Approve()

    expr = pt.Seq(pt.Assert(auth_only(pt.Txn.sender()), comment=cmt), pt.Approve())

    expected = expr.__teal__(options)
    actual = deleter().__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealTypeError):
        Authorize.opted_in(pt.Bytes("abc"))

    with pytest.raises(pt.TealInputError):

        @pt.Subroutine(pt.TealType.uint64)
        def thing(a, b):
            return pt.Int(1)

        @external(authorize=thing)
        def other_thing():
            pass

    with pytest.raises(pt.TealTypeError):

        @pt.Subroutine(pt.TealType.bytes)
        def thing(x):
            return pt.Bytes("fail")

        @external(authorize=thing)
        def other_other_thing():
            pass


def test_named_tuple():
    class Order(pt.abi.NamedTuple):
        item: pt.abi.Field[pt.abi.String]
        count: pt.abi.Field[pt.abi.Uint64]

    @external
    def thing(o: Order):
        pass

    hc = get_handler_config(thing)
    assert hc.structs["o"] is Order


def test_bare():
    @create
    def impl():
        return pt.Assert(pt.Int(1))

    hc = get_handler_config(impl)
    assert hc.bare_method.no_op.action.subroutine.implementation == impl

    @no_op
    def impl():
        return pt.Assert(pt.Int(1))

    hc = get_handler_config(impl)
    assert hc.bare_method.no_op.action.subroutine.implementation == impl

    @delete
    def impl():
        return pt.Assert(pt.Int(1))

    hc = get_handler_config(impl)
    assert hc.bare_method.delete_application.action.subroutine.implementation == impl

    @update
    def impl():
        return pt.Assert(pt.Int(1))

    hc = get_handler_config(impl)
    assert hc.bare_method.update_application.action.subroutine.implementation == impl

    @opt_in
    def impl():
        return pt.Assert(pt.Int(1))

    hc = get_handler_config(impl)
    assert hc.bare_method.opt_in.action.subroutine.implementation == impl

    @close_out
    def impl():
        return pt.Assert(pt.Int(1))

    hc = get_handler_config(impl)
    assert hc.bare_method.close_out.action.subroutine.implementation == impl

    @clear_state
    def impl():
        return pt.Assert(pt.Int(1))

    hc = get_handler_config(impl)
    assert hc.bare_method.clear_state.action.subroutine.implementation == impl


def test_resolvable():
    from .state import (
        AccountStateValue,
        ApplicationStateValue,
        ReservedAccountStateValue,
        ReservedApplicationStateValue,
    )

    x = AccountStateValue(pt.TealType.uint64, key=pt.Bytes("x"))
    r = DefaultArgument(x)
    assert r.resolvable_class == "local-state"

    x = ReservedAccountStateValue(pt.TealType.uint64, max_keys=1)
    r = DefaultArgument(x[pt.Bytes("x")])
    assert r.resolvable_class == "local-state"

    x = ApplicationStateValue(pt.TealType.uint64, key=pt.Bytes("x"))
    r = DefaultArgument(x)
    assert r.resolvable_class == "global-state"

    x = ReservedApplicationStateValue(pt.TealType.uint64, max_keys=1)
    r = DefaultArgument(x[pt.Bytes("x")])
    assert r.resolvable_class == "global-state"

    @external(read_only=True)
    def x():
        return pt.Assert(pt.Int(1))

    r = DefaultArgument(x)
    assert r.resolvable_class == "abi-method"

    r = DefaultArgument(pt.Bytes("1"))
    assert r.resolvable_class == "constant"

    r = DefaultArgument(pt.Int(1))
    assert r.resolvable_class == "constant"



from beaker import *
from .nested_application import *


def demo():
    accts = sandbox.get_accounts()
    acct = accts.pop()

    # Create grandparent app and fund it
    app_client_grandparent = client.ApplicationClient(
        sandbox.get_algod_client(), Grandparent(), signer=acct.signer
    )
    grandparent_app_id, _, _ = app_client_grandparent.create()
    print(f"Created grandparent app: {grandparent_app_id}")
    app_client_grandparent.fund(1 * consts.algo)

    # Call the grandparent app to create the parent app
    result = app_client_grandparent.call(Grandparent.create_parent)
    parent_app_id = result.return_value
    print(f"Created parent app: {parent_app_id}")

    # Create parent app client
    app_client_parent = client.ApplicationClient(
        sandbox.get_algod_client(),
        Parent(),
        signer=acct.signer,
        app_id=parent_app_id,
    )

    app_client_parent.fund(1 * consts.algo)

    # Call the parent app to create the child_1 app
    result = app_client_parent.call(Parent.create_child_1)
    child_app_id = result.return_value
    print(f"Created child_1 app: {child_app_id}")

    # Create child_1 app client
    app_client_child = client.ApplicationClient(
        sandbox.get_algod_client(),
        Child1(),
        signer=acct.signer,
        app_id=child_app_id,
    )

    app_client_child.fund(1 * consts.algo)

    # Call the child_1 app to increment counter
    result = app_client_child.call(Child1.increment_counter)
    counter_value = result.return_value
    print(f"Counter value: {counter_value}")

    # Call the parent app to create the child_2 app
    result = app_client_parent.call(Parent.create_child_2)
    child_app_id = result.return_value
    print(f"Created child_2 app: {child_app_id}")

    # Create child_2 app client
    app_client_child = client.ApplicationClient(
        sandbox.get_algod_client(),
        Child2(),
        signer=acct.signer,
        app_id=child_app_id,
    )

    app_client_child.fund(1 * consts.algo)

    # Call the child_2 app to check lsig addr
    result = app_client_child.call(Child2.get_lsig_addr)
    addr_value = result.return_value
    print(f"LSig address value: {addr_value}")


if __name__ == "__main__":
    demo()



from algosdk.abi import ABIType
from algosdk.encoding import encode_address, decode_address
from algosdk.atomic_transaction_composer import TransactionWithSigner
from algosdk.future.transaction import *

from pyteal import *
from beaker import *

if __name__ == "__main__":
    from application import AppMember, MembershipRecord, MembershipClub
else:
    from .application import AppMember, MembershipRecord, MembershipClub


record_codec = ABIType.from_string(str(MembershipRecord().type_spec()))


affirmations = [
    "I am successful.",
    "I am confident.",
    "I am powerful.",
    "I am strong.",
    "I am getting better and better every day.",
    "All I need is within me right now.",
    "I wake up motivated.",
    "I am an unstoppable force of nature.",
    "I am a living, breathing example of motivation.",
    "All I need is GM.",
]


def print_boxes(app_client: client.ApplicationClient):
    boxes = app_client.get_box_names()
    print(f"{len(boxes)} boxes found")
    for box_name in boxes:
        contents = app_client.get_box_contents(box_name)
        if box_name == b"affirmations":
            print(contents)
        else:
            membership_record = record_codec.decode(contents)
            print(f"\t{encode_address(box_name)} => {membership_record} ")


def demo():
    accts = sandbox.get_accounts()
    acct = accts.pop()
    member_acct = accts.pop()

    app_client = client.ApplicationClient(
        sandbox.get_algod_client(), MembershipClub(), signer=acct.signer
    )
    print("Creating app")
    app_client.create()

    ##
    # Bootstrap Club app
    ##
    print("Bootstrapping app")
    sp = app_client.get_suggested_params()
    sp.flat_fee = True
    sp.fee = 2000
    ptxn = PaymentTxn(
        acct.address, sp, app_client.app_addr, MembershipClub._min_balance
    )
    result = app_client.call(
        MembershipClub.bootstrap,
        seed=TransactionWithSigner(ptxn, acct.signer),
        token_name="fight club",
        boxes=[[app_client.app_id, "affirmations"]] * 8,
    )
    membership_token = result.return_value
    print(f"Created asset id: {membership_token}")

    ##
    # Add Member to club
    ##

    # Opt member account in to asset
    app_client.client.send_transaction(
        AssetOptInTxn(member_acct.address, sp, membership_token).sign(
            member_acct.private_key
        )
    )

    # Add member account as member
    app_client.call(
        MembershipClub.add_member,
        new_member=member_acct.address,
        suggested_params=sp,
        boxes=[[app_client.app_id, decode_address(member_acct.address)]],
    )
    print_boxes(app_client)

    # read the membership record box
    result = app_client.call(
        MembershipClub.get_membership_record,
        member=member_acct.address,
        boxes=[[app_client.app_id, decode_address(member_acct.address)]],
    )
    print(result.return_value)

    # Create a new client for the member
    member_client = app_client.prepare(signer=member_acct.signer)
    for idx, aff in enumerate(affirmations):
        result = member_client.call(
            MembershipClub.set_affirmation,
            idx=idx,
            affirmation=aff.ljust(64, " ").encode(),
            boxes=[[app_client.app_id, "affirmations"]],
        )

    # Get the affirmation from the app, passing box ref holding affirmations
    result = member_client.call(
        MembershipClub.get_affirmation,
        boxes=[[app_client.app_id, "affirmations"]],
    )
    print(bytes(result.return_value).decode("utf-8").strip())

    # Remove the member we'd just added
    app_client.call(
        MembershipClub.remove_member,
        boxes=[[app_client.app_id, decode_address(member_acct.address)]],
        member=member_acct.address,
    )
    print_boxes(app_client)

    ##
    # Create Application that will be a member of the MembershipClub
    ##

    # Create App we'll use to be a member of club
    print("Creating app member")
    app_member_client = client.ApplicationClient(
        sandbox.get_algod_client(), AppMember(), signer=app_client.signer
    )
    _, app_member_addr, _ = app_member_client.create()

    # Fund the app member and make it opt into the membership token
    print("Bootstrapping app member")
    sp = app_member_client.get_suggested_params()
    sp.flat_fee = True
    sp.fee = 2000
    ptxn = PaymentTxn(app_client.sender, sp, app_member_addr, consts.algo * 1)
    app_member_client.call(
        AppMember.bootstrap,
        seed=TransactionWithSigner(ptxn, app_client.signer),
        app_id=app_client.app_id,
        membership_token=membership_token,
    )

    # Add app to club using the member_club client
    app_client.call(
        MembershipClub.add_member,
        new_member=app_member_addr,
        suggested_params=sp,
        boxes=[[app_client.app_id, decode_address(app_member_addr)]],
    )

    # Call method to get a new affirmation
    app_member_client.call(
        AppMember.get_affirmation, boxes=[[app_client.app_id, "affirmations"]]
    )

    # Read the affirmation out of the AppMembers app state
    app_state = app_member_client.get_application_state()
    print(f"Last affirmation received by app member: {app_state['last_affirmation']}")


if __name__ == "__main__":
    demo()



from algosdk.future import transaction
from algosdk.atomic_transaction_composer import (
    AtomicTransactionComposer,
    TransactionWithSigner,
)
from beaker import consts

from beaker.sandbox import get_accounts, get_algod_client
from beaker.client import ApplicationClient

from .amm import ConstantProductAMM


# Take first account from sandbox
acct = get_accounts().pop()
addr, sk, signer = acct.address, acct.private_key, acct.signer

# get sandbox client
client = get_algod_client()

# Create an Application client containing both an algod client and my app
app_client = ApplicationClient(client, ConstantProductAMM(), signer=acct.signer)


def demo():

    # Create the applicatiion on chain, set the app id for the app client
    app_id, app_addr, txid = app_client.create()
    print(f"Created App with id: {app_id} and address addr: {app_addr} in tx: {txid}")

    # Fund App address so it can create the pool token and hold balances

    # Create assets
    asset_a = create_asset(addr, sk, "A")
    asset_b = create_asset(addr, sk, "B")
    print(f"Created asset a/b with ids: {asset_a}/{asset_b}")

    # Call app to create pool token
    print("Calling bootstrap")
    sp = client.suggested_params()
    ptxn = TransactionWithSigner(
        txn=transaction.PaymentTxn(addr, sp, app_addr, int(1e7)), signer=signer
    )
    sp.flat_fee = True
    sp.fee = consts.milli_algo * 4
    result = app_client.call(
        ConstantProductAMM.bootstrap,
        seed=ptxn,
        a_asset=asset_a,
        b_asset=asset_b,
        suggested_params=sp,
    )
    pool_token = result.return_value
    print(f"Created pool token with id: {pool_token}")
    print_balances(app_id, app_addr, addr, pool_token, asset_a, asset_b)

    # Opt user into token
    sp = client.suggested_params()
    atc = AtomicTransactionComposer()
    atc.add_transaction(
        TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, addr, 0, pool_token),
            signer=signer,
        )
    )
    atc.execute(client, 2)
    print_balances(app_id, app_addr, addr, pool_token, asset_a, asset_b)

    # Cover any fees incurred by inner transactions, maybe overpaying but thats ok
    sp = client.suggested_params()
    sp.flat_fee = True
    sp.fee = consts.milli_algo * 3

    ###
    # Fund Pool with initial liquidity
    ###
    print("Funding")
    app_client.call(
        ConstantProductAMM.mint,
        a_xfer=TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, 10000, asset_a),
            signer=signer,
        ),
        b_xfer=TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, 3000, asset_b),
            signer=signer,
        ),
        suggested_params=sp,
    )
    print_balances(app_id, app_addr, addr, pool_token, asset_a, asset_b)

    ###
    # Mint pool tokens
    ###
    print("Minting")
    app_client.call(
        ConstantProductAMM.mint,
        a_xfer=TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, 100000, asset_a),
            signer=signer,
        ),
        b_xfer=TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, 1000, asset_b),
            signer=signer,
        ),
        suggested_params=sp,
    )
    print_balances(app_id, app_addr, addr, pool_token, asset_a, asset_b)

    ###
    # Swap A for B
    ###
    print("Swapping A for B")
    app_client.call(
        ConstantProductAMM.swap,
        swap_xfer=TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, 500, asset_a),
            signer=signer,
        ),
    )
    print_balances(app_id, app_addr, addr, pool_token, asset_a, asset_b)

    ###
    # Swap B for A
    ###
    print("Swapping B for A")
    app_client.call(
        ConstantProductAMM.swap,
        swap_xfer=TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, 500, asset_b),
            signer=signer,
        ),
    )
    print_balances(app_id, app_addr, addr, pool_token, asset_a, asset_b)

    ###
    # Burn pool tokens
    ###
    print("Burning")
    app_client.call(
        ConstantProductAMM.burn,
        pool_xfer=TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, 100, pool_token),
            signer=signer,
        ),
    )
    print_balances(app_id, app_addr, addr, pool_token, asset_a, asset_b)


def create_asset(addr, pk, unitname):
    # Get suggested params from network
    sp = client.suggested_params()
    # Create the transaction
    create_txn = transaction.AssetCreateTxn(
        addr, sp, 1000000, 0, False, asset_name="asset", unit_name=unitname
    )
    # Ship it
    txid = client.send_transaction(create_txn.sign(pk))
    # Wait for the result so we can return the app id
    result = transaction.wait_for_confirmation(client, txid, 4)
    return result["asset-index"]


def print_balances(app_id: int, app: str, addr: str, pool: int, a: int, b: int):

    addrbal = client.account_info(addr)
    print("Participant: ")
    for asset in addrbal["assets"]:
        if asset["asset-id"] == pool:
            print("\tPool Balance {}".format(asset["amount"]))
        if asset["asset-id"] == a:
            print("\tAssetA Balance {}".format(asset["amount"]))
        if asset["asset-id"] == b:
            print("\tAssetB Balance {}".format(asset["amount"]))

    appbal = client.account_info(app)
    print("App: ")
    for asset in appbal["assets"]:
        if asset["asset-id"] == pool:
            print("\tPool Balance {}".format(asset["amount"]))
        if asset["asset-id"] == a:
            print("\tAssetA Balance {}".format(asset["amount"]))
        if asset["asset-id"] == b:
            print("\tAssetB Balance {}".format(asset["amount"]))

    state = app_client.get_application_state()
    state_key = ConstantProductAMM.ratio.str_key()
    if state_key in state:
        print(
            f"\tCurrent ratio a/b == {int(state[state_key]) / ConstantProductAMM._scale}"
        )
    else:
        print("\tNo ratio a/b")


if __name__ == "__main__":
    demo()



from pyteal import *
from beaker import *

# Create a class, subclassing Application from beaker
class HelloBeaker(Application):
    # Add an external method with ABI method signature `hello(string)string`
    @external
    def hello(self, name: abi.String, *, output: abi.String):
        # Set output to the result of `Hello, `+name
        return output.set(Concat(Bytes("Hello, "), name.get()))


def demo():
    # Create an Application client
    app_client = client.ApplicationClient(
        # Get sandbox algod client
        client=sandbox.get_algod_client(),
        # Instantiate app with the program version (default is MAX_TEAL_VERSION)
        app=HelloBeaker(version=6),
        # Get acct from sandbox and pass the signer
        signer=sandbox.get_accounts().pop().signer,
    )

    # Deploy the app on-chain
    app_id, app_addr, txid = app_client.create()
    print(
        f"""Deployed app in txid {txid}
        App ID: {app_id} 
        Address: {app_addr} 
    """
    )

    # Call the `hello` method
    result = app_client.call(HelloBeaker.hello, name="Beaker")
    print(result.return_value)  # "Hello, Beaker"


if __name__ == "__main__":
    demo()



from .calculator import demo as calc_demo
from .counter import demo as count_demo
from .hello import demo as hello_demo


def test_calc():
    calc_demo()


def test_count():
    count_demo()


def test_hello():
    hello_demo()



import pytest
import pyteal as pt
import beaker as bkr

from beaker.testing import (
    UnitTestingApp,
    assert_output,
    returned_int_as_bytes,
)

from beaker.lib.strings import atoi, encode_uvarint, head, itoa, prefix, suffix, tail


def test_atoi():
    ut = UnitTestingApp(pt.Itob(atoi(pt.Bytes("123"))))
    output = [returned_int_as_bytes(int(123))]
    assert_output(ut, [], output)


def test_atoi_invalid():
    ut = UnitTestingApp(pt.Itob(atoi(pt.Bytes("abc"))))
    output = [returned_int_as_bytes(int(123))]
    with pytest.raises(bkr.client.LogicException):
        assert_output(ut, [], output)


def test_itoa():
    ut = UnitTestingApp(itoa(pt.Int(123)))
    output = [list(b"123")]
    assert_output(ut, [], output)


def test_head():
    ut = UnitTestingApp(head(pt.Bytes("deadbeef")))
    output = [list(b"d")]
    assert_output(ut, [], output)


def test_head_empty():
    ut = UnitTestingApp(tail(pt.Bytes("")))

    with pytest.raises(bkr.client.LogicException):
        assert_output(ut, [], [None])


def test_tail():
    ut = UnitTestingApp(tail(pt.Bytes("deadbeef")))
    output = [list(b"eadbeef")]
    assert_output(ut, [], output)


def test_tail_empty():
    ut = UnitTestingApp(tail(pt.Bytes("")))
    with pytest.raises(bkr.client.LogicException):
        assert_output(ut, [], [None])


def test_suffix():
    ut = UnitTestingApp(suffix(pt.Bytes("deadbeef"), pt.Int(2)))
    output = [list(b"ef")]
    assert_output(ut, [], output)


def test_suffix_past_length():
    ut = UnitTestingApp(suffix(pt.Bytes("deadbeef"), pt.Int(9)))
    with pytest.raises(bkr.client.LogicException):
        assert_output(ut, [], [None])


def test_prefix():
    ut = UnitTestingApp(prefix(pt.Bytes("deadbeef"), pt.Int(2)))
    output = [list(b"de")]
    assert_output(ut, [], output)


def test_prefix_past_length():
    ut = UnitTestingApp(prefix(pt.Bytes("deadbeef"), pt.Int(9)))
    with pytest.raises(bkr.client.LogicException):
        assert_output(ut, [], [None])


def test_encode_uvarint():
    ut = UnitTestingApp(encode_uvarint(pt.Int(500)))
    output = [[244, 3]]
    assert_output(ut, [], output)



from pyteal import abi

from beaker.client import ApplicationClient
from beaker.application import Application
from beaker.decorators import external
from beaker import sandbox


class Calculator(Application):
    @external
    def add(self, a: abi.Uint64, b: abi.Uint64, *, output: abi.Uint64):
        """Add a and b, return the result"""
        return output.set(a.get() + b.get())

    @external
    def mul(self, a: abi.Uint64, b: abi.Uint64, *, output: abi.Uint64):
        """Multiply a and b, return the result"""
        return output.set(a.get() * b.get())

    @external
    def sub(self, a: abi.Uint64, b: abi.Uint64, *, output: abi.Uint64):
        """Subtract b from a, return the result"""
        return output.set(a.get() - b.get())

    @external
    def div(self, a: abi.Uint64, b: abi.Uint64, *, output: abi.Uint64):
        """Divide a by b, return the result"""
        return output.set(a.get() / b.get())


def demo():
    # Here we use `sandbox` but beaker.client.api_providers can also be used
    # with something like ``AlgoNode(Network.TestNet).algod()``
    algod_client = sandbox.get_algod_client()

    acct = sandbox.get_accounts().pop()

    # Create an Application client containing both an algod client and app
    app_client = ApplicationClient(
        client=algod_client, app=Calculator(), signer=acct.signer
    )

    # Create the application on chain, set the app id for the app client
    app_id, app_addr, txid = app_client.create()
    print(f"Created App with id: {app_id} and address addr: {app_addr} in tx: {txid}")

    result = app_client.call(Calculator.add, a=2, b=2)
    print(f"add result: {result.return_value}")

    result = app_client.call(Calculator.mul, a=2, b=2)
    print(f"mul result: {result.return_value}")

    result = app_client.call(Calculator.sub, a=6, b=2)
    print(f"sub result: {result.return_value}")

    result = app_client.call(Calculator.div, a=16, b=4)
    print(f"div result: {result.return_value}")


if __name__ == "__main__":
    import json

    calc = Calculator()
    print(calc.approval_program)
    print(calc.clear_program)
    print(json.dumps(calc.contract.dictify()))

    demo()



import base64
from copy import copy
from nacl.signing import SigningKey

from algosdk.encoding import decode_address
from algosdk.future import transaction
from algosdk.atomic_transaction_composer import (
    AtomicTransactionComposer,
    TransactionWithSigner,
)

from typing import Literal
from pyteal import *
from beaker import *
from beaker.precompile import LSigPrecompile

Signature = abi.StaticBytes[Literal[64]]


class App(Application):
    class SigChecker(LogicSignature):
        # Simple program to check an ed25519 signature given a message and signature
        user_addr = TemplateVariable(stack_type=TealType.bytes)

        def evaluate(self):
            return Seq(
                # Borrow the msg and sig from the abi call arguments
                # TODO: this kinda stinks, what do?
                (msg := abi.String()).decode(Txn.application_args[2]),
                (sig := abi.make(Signature)).decode(Txn.application_args[3]),
                # Assert that the sig matches
                Assert(Ed25519Verify_Bare(msg.get(), sig.get(), self.user_addr)),
                Int(1),
            )

    sig_checker = LSigPrecompile(SigChecker())

    @external
    def check(self, signer_address: abi.Address, msg: abi.String, sig: Signature):
        return Assert(
            Txn.sender() == self.sig_checker.logic.template_hash(signer_address.get())
        )


def sign_msg(msg: str, sk: str) -> bytes:
    """utility function for signing arbitrary data"""
    pk: list[bytes] = list(base64.b64decode(sk))
    return SigningKey(bytes(pk[:32])).sign(msg.encode()).signature


def demo():
    acct = sandbox.get_accounts().pop()

    # Create app client
    app = App()
    app_client = client.ApplicationClient(
        sandbox.get_algod_client(), app, signer=acct.signer
    )

    # deploy app
    app_client.create()

    # Write the populated template as binary
    # with open("tmp.teal.tok", "wb") as f:
    #     f.write(app.sig_checker.populate_template(decode_address(acct.address)))

    # Get the signer for the lsig from its populated precompile
    lsig_signer = app.sig_checker.template_signer(decode_address(acct.address))
    # Prepare a new client so it can sign calls
    lsig_client = app_client.prepare(signer=lsig_signer)

    atc = AtomicTransactionComposer()

    # Create a dummy transaction to cover fees (still cheaper than multiple app calls)
    sp = app_client.client.suggested_params()
    covering_sp = copy(sp)
    covering_sp.flat_fee = True
    covering_sp.fee = 2 * consts.milli_algo
    atc.add_transaction(
        TransactionWithSigner(
            txn=transaction.PaymentTxn(acct.address, covering_sp, acct.address, 0),
            signer=acct.signer,
        )
    )

    # Message to sign
    msg = "Sign me please"
    # Signature
    sig = sign_msg(msg, acct.signer.private_key)

    # Since we dont fund this account, just set its fees to 0
    free_sp = copy(sp)
    free_sp.flat_fee = True
    free_sp.fee = 0

    # Add the call to the `check` method to be signed by the populated template logic
    lsig_client.add_method_call(
        atc,
        App.check,
        suggested_params=free_sp,
        signer_address=acct.address,
        msg=msg,
        sig=sig,
    )

    # run it
    result = atc.execute(app_client.client, 4)
    print(f"Confirmed in round {result.confirmed_round}")


if __name__ == "__main__":
    demo()



import pytest
import pyteal as pt
from typing import Any
from base64 import b64decode, b64encode

from algosdk.account import generate_account
from algosdk.logic import get_application_address
from algosdk.future.transaction import Multisig, LogicSigAccount, OnComplete
from algosdk.atomic_transaction_composer import (
    AtomicTransactionComposer,
    AccountTransactionSigner,
    MultisigTransactionSigner,
    LogicSigTransactionSigner,
)

from beaker.decorators import (
    Authorize,
    DefaultArgument,
    create,
    external,
    update,
    clear_state,
    close_out,
    delete,
    opt_in,
)
from beaker.sandbox import get_accounts, get_algod_client
from beaker.application import Application, get_method_selector
from beaker.state import ApplicationStateValue, AccountStateValue
from beaker.client.application_client import ApplicationClient
from beaker.client.logic_error import LogicException


class App(Application):
    app_state_val_int = ApplicationStateValue(pt.TealType.uint64, default=pt.Int(1))
    app_state_val_byte = ApplicationStateValue(
        pt.TealType.bytes, default=pt.Bytes("test")
    )
    acct_state_val_int = AccountStateValue(pt.TealType.uint64, default=pt.Int(1))
    acct_state_val_byte = AccountStateValue(pt.TealType.bytes, default=pt.Bytes("test"))

    @create
    def create(self):
        return pt.Seq(
            self.initialize_application_state(),
            pt.Assert(pt.Len(pt.Txn.note()) == pt.Int(0)),
            pt.Approve(),
        )

    @update(authorize=Authorize.only(pt.Global.creator_address()))
    def update(self):
        return pt.Approve()

    @delete(authorize=Authorize.only(pt.Global.creator_address()))
    def delete(self):
        return pt.Approve()

    @opt_in
    def opt_in(self):
        return pt.Seq(
            self.initialize_account_state(),
            pt.Assert(pt.Len(pt.Txn.note()) == pt.Int(0)),
            pt.Approve(),
        )

    @clear_state
    def clear_state(self):
        return pt.Seq(pt.Assert(pt.Len(pt.Txn.note()) == pt.Int(0)), pt.Approve())

    @close_out
    def close_out(self):
        return pt.Seq(pt.Assert(pt.Len(pt.Txn.note()) == pt.Int(0)), pt.Approve())

    @external
    def add(self, a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64):
        return output.set(a.get() + b.get())

    @external(read_only=True)
    def dummy(self, *, output: pt.abi.String):
        return output.set("deadbeef")


SandboxAccounts = list[tuple[str, str, AccountTransactionSigner]]


@pytest.fixture(scope="session")
def sb_accts() -> SandboxAccounts:
    return [(acct.address, acct.private_key, acct.signer) for acct in get_accounts()]


def test_app_client_create():
    app = App()
    client = get_algod_client()
    ac = ApplicationClient(client, app)
    assert ac.signer is None, "Should not have a signer"
    assert ac.sender is None, "Should not have a sender"
    assert ac.app_id == 0, "Should not have app id"
    assert ac.app_addr is None, "Should not have app address"
    assert ac.suggested_params is None, "Should not have suggested params"

    with pytest.raises(Exception):
        ac.get_signer(None)

    with pytest.raises(Exception):
        ac.get_sender(None, None)


def test_app_prepare(sb_accts: SandboxAccounts):
    app = App()
    client = get_algod_client()

    (addr, sk, signer) = sb_accts[0]

    ac_with_signer = ApplicationClient(client, app, signer=signer)

    assert ac_with_signer.signer == signer, "Should have the same signer"

    assert ac_with_signer.get_signer(None) == signer, "Should produce the same signer"
    assert (
        ac_with_signer.get_sender(None, None) == addr
    ), "Should produce the same address"

    new_pk, new_addr = generate_account()
    new_signer = AccountTransactionSigner(new_pk)
    ac_with_signer_and_sender = ac_with_signer.prepare(sender=new_addr)

    assert (
        ac_with_signer_and_sender.signer == signer
    ), "Should have the same original signer"
    assert (
        ac_with_signer_and_sender.sender == new_addr
    ), "Should have the new addr as sender"

    assert (
        ac_with_signer_and_sender.get_signer(None) == signer
    ), "Should produce the same signer"

    assert (
        ac_with_signer_and_sender.get_sender(None, None) == new_addr
    ), "Should produce the new address"

    assert (
        ac_with_signer_and_sender.get_signer(new_signer) == new_signer
    ), "Should be new signer"
    assert (
        ac_with_signer_and_sender.get_sender(None, new_signer) == new_addr
    ), "Should be new address"

    accts = [generate_account() for _ in range(3)]
    addrs = [acct[1] for acct in accts]
    sks = [acct[0] for acct in accts]

    msig_acct = Multisig(1, 3, addrs)
    msts = MultisigTransactionSigner(msig_acct, sks[0])

    ac_with_msig = ac_with_signer.prepare(signer=msts)
    assert ac_with_msig.signer == msts, "Should have the same signer"
    assert (
        ac_with_msig.sender == msig_acct.address()
    ), "Should have the address of the msig as the sender"
    assert ac_with_msig.get_signer(None) == msts, "Should produce the same signer"
    assert (
        ac_with_msig.get_sender(None, None) == msig_acct.address()
    ), "Should produce the same address"

    # pragma version 6; int 1; return
    program = b64decode("BoEBQw==")
    lsig = LogicSigAccount(program)
    lsig_signer = LogicSigTransactionSigner(lsig)

    ac_with_lsig = ac_with_signer.prepare(signer=lsig_signer)
    assert ac_with_lsig.signer == lsig_signer, "Should have the same signer"
    assert (
        ac_with_lsig.sender == lsig.address()
    ), "Should have the address of the lsig as the sender"
    assert (
        ac_with_lsig.get_signer(None) == lsig_signer
    ), "Should produce the same signer"
    assert (
        ac_with_lsig.get_sender(None, None) == lsig.address()
    ), "Should produce the same address"

    ac_with_app_id = ac_with_signer.prepare(app_id=3)
    assert (
        ac_with_signer.app_id == 0
    ), "We should not have changed the app id in the original"
    assert (
        ac_with_app_id.app_id == 3
    ), "We should have overwritten the app id in the new version"


def test_compile():
    version = 5
    app = App(version=version)
    client = get_algod_client()
    ac = ApplicationClient(client, app)

    # TODO add precompiles

    approval_program, _, approval_map = ac.compile(
        ac.app.approval_program, source_map=True
    )

    assert len(approval_program) > 0, "Should have a valid approval program"
    assert approval_program[0] == version, "First byte should be the version we set"
    assert approval_map.version == 3, "Should have valid source map with version 3"
    assert len(approval_map.pc_to_line) > 0, "Should have valid mapping"

    clear_program, _, clear_map = ac.compile(ac.app.clear_program, source_map=True)
    assert len(clear_program) > 0, "Should have a valid clear program"
    assert clear_program[0] == version, "First byte should be the version we set"
    assert clear_map.version == 3, "Should have valid source map with version 3"
    assert len(clear_map.pc_to_line) > 0, "Should have valid mapping"


def expect_dict(actual: dict[str, Any], expected: dict[str, Any]):
    for k, v in expected.items():
        if type(v) is dict:
            expect_dict(actual[k], v)
        else:
            assert actual[k] == v, f"for field {k}, expected {v} got {actual[k]}"


def test_create(sb_accts: SandboxAccounts):
    app = App()

    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, app_addr, tx_id = ac.create()
    assert app_id > 0
    assert app_addr == get_application_address(app_id)
    assert ac.app_id == app_id
    assert ac.app_addr == app_addr

    result_tx = client.pending_transaction_info(tx_id)
    assert result_tx["confirmed-round"] > 0
    expect_dict(
        result_tx,
        {
            "application-index": app_id,
            "pool-error": "",
            "txn": {
                "txn": {
                    "snd": addr,
                    "apgs": {"nbs": 1, "nui": 1},
                    "apls": {"nbs": 1, "nui": 1},
                }
            },
        },
    )

    new_addr, new_pk, new_signer = sb_accts[1]
    new_ac = ac.prepare(signer=new_signer)
    extra_pages = 2
    sp = client.suggested_params()
    sp.fee = 1_000_000
    sp.flat_fee = True
    app_id, app_addr, tx_id = new_ac.create(
        extra_pages=extra_pages, suggested_params=sp
    )
    assert app_id > 0
    assert app_addr == get_application_address(app_id)
    assert new_ac.app_id == app_id
    assert new_ac.app_addr == app_addr

    result_tx = client.pending_transaction_info(tx_id)
    expect_dict(
        result_tx,
        {
            "application-index": app_id,
            "pool-error": "",
            "txn": {
                "txn": {
                    "snd": new_addr,
                    "apep": extra_pages,
                    "fee": sp.fee,
                    "apgs": {"nbs": 1, "nui": 1},
                    "apls": {"nbs": 1, "nui": 1},
                }
            },
        },
    )

    with pytest.raises(LogicException):
        ac.create(note="failmeplz")


def test_update(sb_accts: SandboxAccounts):
    app = App()

    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, app_addr, _ = ac.create()

    tx_id = ac.update()
    result_tx = client.pending_transaction_info(tx_id)
    expect_dict(
        result_tx,
        {
            "pool-error": "",
            "txn": {
                "txn": {
                    "apan": OnComplete.UpdateApplicationOC,
                    "apid": app_id,
                    "snd": addr,
                }
            },
        },
    )

    with pytest.raises(LogicException):
        addr, pk, signer2 = sb_accts[1]
        ac2 = ac.prepare(signer=signer2)
        ac2.update()


def test_delete(sb_accts: SandboxAccounts):
    app = App()
    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, _, _ = ac.create()

    tx_id = ac.delete()
    result_tx = client.pending_transaction_info(tx_id)
    expect_dict(
        result_tx,
        {
            "pool-error": "",
            "txn": {
                "txn": {
                    "apan": OnComplete.DeleteApplicationOC,
                    "apid": app_id,
                    "snd": addr,
                }
            },
        },
    )

    with pytest.raises(LogicException):
        ac = ApplicationClient(client, app, signer=signer)
        app_id, _, _ = ac.create()

        _, _, signer2 = sb_accts[1]
        ac2 = ac.prepare(signer=signer2)
        ac2.delete()


def test_opt_in(sb_accts: SandboxAccounts):
    app = App()

    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, _, _ = ac.create()

    new_addr, new_pk, new_signer = sb_accts[1]
    new_ac = ac.prepare(signer=new_signer)
    tx_id = new_ac.opt_in()
    result_tx = client.pending_transaction_info(tx_id)
    expect_dict(
        result_tx,
        {
            "pool-error": "",
            "txn": {
                "txn": {
                    "apan": OnComplete.OptInOC,
                    "apid": app_id,
                    "snd": new_addr,
                }
            },
        },
    )

    with pytest.raises(LogicException):
        _, _, newer_signer = sb_accts[2]
        newer_ac = ac.prepare(signer=newer_signer)
        newer_ac.opt_in(note="failmeplz")


def test_close_out(sb_accts: SandboxAccounts):

    app = App()

    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, _, _ = ac.create()

    new_addr, new_pk, new_signer = sb_accts[1]
    new_ac = ac.prepare(signer=new_signer)
    new_ac.opt_in()

    tx_id = new_ac.close_out()
    result_tx = client.pending_transaction_info(tx_id)
    expect_dict(
        result_tx,
        {
            "pool-error": "",
            "txn": {
                "txn": {
                    "apan": OnComplete.CloseOutOC,
                    "apid": app_id,
                    "snd": new_addr,
                }
            },
        },
    )

    with pytest.raises(LogicException):
        _, _, newer_signer = sb_accts[2]
        newer_ac = ac.prepare(signer=newer_signer)
        newer_ac.opt_in()
        newer_ac.close_out(note="failmeplz")


def test_clear_state(sb_accts: SandboxAccounts):
    app = App()
    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, _, _ = ac.create()

    new_addr, new_pk, new_signer = sb_accts[1]
    new_ac = ac.prepare(signer=new_signer)
    new_ac.opt_in()

    tx_id = new_ac.clear_state()
    result_tx = client.pending_transaction_info(tx_id)
    expect_dict(
        result_tx,
        {
            "pool-error": "",
            "txn": {
                "txn": {
                    "apan": OnComplete.ClearStateOC,
                    "apid": app_id,
                    "snd": new_addr,
                }
            },
        },
    )


def test_call(sb_accts: SandboxAccounts):
    app = App()
    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, _, _ = ac.create()

    result = ac.call(app.add, a=1, b=1)
    assert result.return_value == 2
    assert result.decode_error is None
    assert result.raw_value == (2).to_bytes(8, "big")

    ms = get_method_selector(app.add)
    raw_args = [ms, (1).to_bytes(8, "big"), (1).to_bytes(8, "big")]

    return_prefix = 0x151F7C75
    log_msg = b64encode(
        return_prefix.to_bytes(4, "big") + (2).to_bytes(8, "big")
    ).decode("utf-8")

    result_tx = client.pending_transaction_info(result.tx_id)
    expect_dict(
        result_tx,
        {
            "pool-error": "",
            "txn": {
                "txn": {
                    "apaa": [b64encode(arg).decode("utf-8") for arg in raw_args],
                    "apid": app_id,
                    "snd": addr,
                }
            },
            "logs": [log_msg],
        },
    )


def test_add_method_call(sb_accts: SandboxAccounts):
    app = App()

    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)
    app_id, _, _ = ac.create()

    atc = AtomicTransactionComposer()
    ac.add_method_call(atc, app.add, a=1, b=1)
    atc_result = atc.execute(client, 4)
    result = atc_result.abi_results[0]

    assert result.return_value == 2
    assert result.decode_error is None
    assert result.raw_value == (2).to_bytes(8, "big")

    ms = get_method_selector(app.add)
    raw_args = [ms, (1).to_bytes(8, "big"), (1).to_bytes(8, "big")]

    return_prefix = 0x151F7C75
    log_msg = b64encode(
        return_prefix.to_bytes(4, "big") + (2).to_bytes(8, "big")
    ).decode("utf-8")

    result_tx = client.pending_transaction_info(result.tx_id)
    expect_dict(
        result_tx,
        {
            "pool-error": "",
            "txn": {
                "txn": {
                    "apaa": [b64encode(arg).decode("utf-8") for arg in raw_args],
                    "apid": app_id,
                    "snd": addr,
                }
            },
            "logs": [log_msg],
        },
    )


def test_fund(sb_accts: SandboxAccounts):
    app = App()
    addr, pk, signer = sb_accts[0]
    client = get_algod_client()

    fund_amt = 1_000_000

    ac = ApplicationClient(client, app, signer=signer)
    ac.create()
    ac.fund(fund_amt)

    info = ac.get_application_account_info()
    assert info["amount"] == fund_amt, "Expected balance to equal fund_amt"


def test_resolve(sb_accts: SandboxAccounts):

    app = App()

    addr, pk, signer = sb_accts[0]

    client = get_algod_client()
    ac = ApplicationClient(client, app, signer=signer)

    ac.create()
    ac.opt_in()

    assert ac.resolve(DefaultArgument(pt.Int(1))) == 1
    assert ac.resolve(DefaultArgument(pt.Bytes("stringy"))) == "stringy"
    assert ac.resolve(DefaultArgument(app.app_state_val_int)) == 1
    assert ac.resolve(DefaultArgument(app.app_state_val_byte)) == b"test"
    assert ac.resolve(DefaultArgument(app.acct_state_val_int)) == 1
    assert ac.resolve(DefaultArgument(app.acct_state_val_byte)) == b"test"
    assert ac.resolve(DefaultArgument(app.dummy)) == "deadbeef"



from dataclasses import dataclass
from algosdk.atomic_transaction_composer import AccountTransactionSigner
from algosdk.kmd import KMDClient

DEFAULT_KMD_ADDRESS = "http://localhost:4002"
DEFAULT_KMD_TOKEN = "a" * 64
DEFAULT_KMD_WALLET_NAME = "unencrypted-default-wallet"
DEFAULT_KMD_WALLET_PASSWORD = ""


@dataclass
class SandboxAccount:
    """SandboxAccount is a simple dataclass to hold a sandbox account details"""

    #: The address of a sandbox account
    address: str
    #: The base64 encoded private key of the account
    private_key: str
    #: An AccountTransactionSigner that can be used as a TransactionSigner
    signer: AccountTransactionSigner


def get_accounts(
    kmd_address: str = DEFAULT_KMD_ADDRESS,
    kmd_token: str = DEFAULT_KMD_TOKEN,
    wallet_name: str = DEFAULT_KMD_WALLET_NAME,
    wallet_password: str = DEFAULT_KMD_WALLET_PASSWORD,
) -> list[SandboxAccount]:
    """gets all the accounts in the sandbox kmd, defaults to the `unencrypted-default-wallet` created on private networks automatically"""

    kmd = KMDClient(kmd_token, kmd_address)
    wallets = kmd.list_wallets()

    wallet_id = None
    for wallet in wallets:
        if wallet["name"] == wallet_name:
            wallet_id = wallet["id"]
            break

    if wallet_id is None:
        raise Exception("Wallet not found: {}".format(wallet_name))

    wallet_handle = kmd.init_wallet_handle(wallet_id, wallet_password)

    try:
        addresses = kmd.list_keys(wallet_handle)
        private_keys = [
            kmd.export_key(wallet_handle, wallet_password, addr) for addr in addresses
        ]
        kmd_accounts = [
            SandboxAccount(
                addresses[i], private_keys[i], AccountTransactionSigner(private_keys[i])
            )
            for i in range(len(private_keys))
        ]
    finally:
        kmd.release_wallet_handle(wallet_handle)

    return kmd_accounts


def add_account(
    private_key: str,
    kmd_address: str = DEFAULT_KMD_ADDRESS,
    kmd_token: str = DEFAULT_KMD_TOKEN,
    wallet_name: str = DEFAULT_KMD_WALLET_NAME,
    wallet_password: str = DEFAULT_KMD_WALLET_PASSWORD,
) -> str:
    """Adds a new account to the sandbox kmd"""

    kmd = KMDClient(kmd_token, kmd_address)
    wallets = kmd.list_wallets()

    wallet_id = None
    for wallet in wallets:
        if wallet["name"] == wallet_name:
            wallet_id = wallet["id"]
            break

    if wallet_id is None:
        raise Exception("Wallet not found: {}".format(wallet_name))

    wallet_handle = kmd.init_wallet_handle(wallet_id, wallet_password)

    try:
        added = kmd.import_key(wallet_handle, private_key)
    finally:
        kmd.release_wallet_handle(wallet_handle)

    return added


def delete_account(
    address: str,
    kmd_address: str = DEFAULT_KMD_ADDRESS,
    kmd_token: str = DEFAULT_KMD_TOKEN,
    wallet_name: str = DEFAULT_KMD_WALLET_NAME,
    wallet_password: str = DEFAULT_KMD_WALLET_PASSWORD,
):
    """Deletes an existing account from the sandbox kmd"""

    kmd = KMDClient(kmd_token, kmd_address)
    wallets = kmd.list_wallets()

    wallet_id = None
    for wallet in wallets:
        if wallet["name"] == wallet_name:
            wallet_id = wallet["id"]
            break

    if wallet_id is None:
        raise Exception("Wallet not found: {}".format(wallet_name))

    wallet_handle = kmd.init_wallet_handle(wallet_id, wallet_password)

    try:
        kmd.delete_key(wallet_handle, password=wallet_password, address=address)
    finally:
        kmd.release_wallet_handle(wallet_handle)



import base64
from Cryptodome.Hash import SHA512
from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING
from pyteal import (
    Seq,
    Bytes,
    Expr,
    ScratchVar,
    TealType,
    TealTypeError,
    TealInputError,
    Int,
    Concat,
    Len,
    Substring,
    Suffix,
    Subroutine,
    Sha512_256,
    TxnField,
    TxnType,
)
from algosdk.v2client.algod import AlgodClient
from algosdk.source_map import SourceMap
from algosdk.future.transaction import LogicSigAccount
from algosdk.atomic_transaction_composer import LogicSigTransactionSigner
from beaker.consts import PROGRAM_DOMAIN_SEPARATOR, num_extra_program_pages
from algosdk.constants import APP_PAGE_MAX_SIZE
from beaker.lib.strings import encode_uvarint

if TYPE_CHECKING:
    from beaker.application import Application
    from beaker.logic_signature import LogicSignature


#: The prefix for template variables that should be substituted
TMPL_PREFIX = "TMPL_"

#: The opcode that should be present just before the byte template variable
PUSH_BYTES = "pushbytes"
#: The opcode that should be present just before the uint64 template variable
PUSH_INT = "pushint"

#: The zero value for byte type
ZERO_BYTES = '""'
#: The zero value for uint64 type
ZERO_INT = "0"


@dataclass
class PrecompileTemplateValue:
    #: The name of the template variable
    name: str = field(kw_only=True)
    #: Whether or not this variable is bytes (if false, its uint64)
    is_bytes: bool = field(kw_only=True)
    #: The line number in the source TEAL this variable is present
    line: int = field(kw_only=True)
    #: The pc of the variable in the assembled bytecode
    pc: int = 0


@dataclass
class ProgramPage:
    #: index of the program page
    index: int = field(kw_only=True, init=True)
    #: binary of the page
    _binary: bytes = field(kw_only=True, init=True)
    #: bytes of the page as pyteal Bytes
    binary: Bytes = field(init=False)
    #: hash of the page in native bytes
    _hash_digest: bytes = field(kw_only=True, init=True)
    #: hash of the page as pyteal Addr
    hash_digest: Bytes = field(init=False)

    def __post_init__(self):
        self.binary = Bytes(self._binary)
        self.hash_digest = Bytes(self._hash_digest)


class Precompile:
    """
    Precompile takes a TEAL program and handles its compilation. Used by AppPrecompile
    and LSigPrecompile for Applications and Logic Signature programs, respectively.
    """

    _program: str = ""
    _binary: Optional[bytes] = None
    _program_hash: Optional[str] = None
    _map: Optional[SourceMap] = None
    _template_values: list[PrecompileTemplateValue] = []
    program_pages: list[ProgramPage]
    binary: Bytes = Bytes("")

    def __init__(self, program: str):
        self._program = program
        lines = self._program.splitlines()
        template_values: list[PrecompileTemplateValue] = []
        # Replace the teal program TMPL_* template variables with
        # the 0 value for the given type and save the list of TemplateValues
        for idx, line in enumerate(lines):
            if TMPL_PREFIX in line:
                op, name, _, _ = line.split(" ")
                tv = PrecompileTemplateValue(
                    name=name[len(TMPL_PREFIX) :], is_bytes=op == PUSH_BYTES, line=idx
                )
                lines[idx] = line.replace(name, ZERO_BYTES if tv.is_bytes else ZERO_INT)
                template_values.append(tv)

        program = "\n".join(lines)

        self._program = program
        self._template_values = template_values

    def assemble(self, client: AlgodClient):
        """Fully compile the program source to binary and generate a source map for matching pc to line number"""
        result = client.compile(self._program, source_map=True)
        self._binary = base64.b64decode(result["result"])
        self._program_hash = result["hash"]

        self._map = SourceMap(result["sourcemap"])

        self._asserts = _gather_asserts(self._program, self._map)

        self.binary = Bytes(self._binary)
        for tv in self._template_values:
            # +1 to acount for the pushbytes/pushint op
            tv.pc = self._map.get_pcs_for_line(tv.line)[0] + 1

        def _hash_program(data: bytes) -> bytes:
            """compute the hash"""
            chksum = SHA512.new(truncate="256")
            chksum.update(PROGRAM_DOMAIN_SEPARATOR.encode() + data)
            return chksum.digest()

        self.program_pages = [
            ProgramPage(
                index=i,
                _binary=self._binary[i : i + APP_PAGE_MAX_SIZE],
                _hash_digest=_hash_program(self._binary[i : i + APP_PAGE_MAX_SIZE]),
            )
            for i in range(0, len(self._binary), APP_PAGE_MAX_SIZE)
        ]

    def hash(self, page_idx: int = 0) -> Expr:
        """hash returns an expression for this Precompile.  It will fail if any template_values are set.

        Args:
            page_idx(optional): If the application has multiple pages, the index of the page can be specified to get the program hash for that page.

        """
        assert self._binary is not None
        assert len(self._template_values) == 0
        if self._program_hash is None:
            raise TealInputError("No address defined for precompile")

        return self.program_pages[page_idx].hash_digest

    def populate_template(self, *args) -> bytes:
        """
        populate_template returns the bytes resulting from patching the set of
        arguments passed into the blank binary

        The args passed should be of the same type and in the same order as the
        template values declared.
        """

        assert self._binary is not None
        assert len(self._template_values) > 0
        assert len(args) == len(self._template_values)

        # Get a copy of the binary so we can work on it in place
        populated_binary = list(self._binary).copy()
        # Any time we add bytes, we need to update the offset so the rest
        # of the pc values can be updated to account for the difference
        offset = 0
        for idx, tv in enumerate(self._template_values):
            arg: str | bytes | int = args[idx]

            if tv.is_bytes:
                if type(arg) is int:
                    raise TealTypeError(type(arg), bytes | str)

                if type(arg) is str:
                    arg = arg.encode("utf-8")

                assert type(arg) is bytes

                # Bytes are encoded as uvarint(len(bytes)) + bytes
                curr_val = py_encode_uvarint(len(arg)) + arg
            else:
                if type(arg) is not int:
                    raise TealTypeError(type(arg), int)
                # Ints are just the uvarint encoded number
                curr_val = py_encode_uvarint(arg)

            # update the working buffer to include the new value, replacing the current 0 value
            populated_binary[tv.pc + offset : tv.pc + offset + 1] = curr_val

            # update the offset with the length(value) - 1 to account for the existing 0 value
            # and help keep track of how to shift the pc later
            offset += len(curr_val) - 1

        return bytes(populated_binary)

    def populate_template_expr(self, *args: Expr) -> Expr:
        """
        populate_template_expr returns the Expr that will patch a
        blank binary given a set of arguments.

        It is called by ``template_address`` to return a Expr that
        can be used to compare with a sender given some arguments.
        """

        # To understand how this works, first look at the pure python one above
        # it should produce an identical output in terms of populated binary.
        # This function just reproduces the same effects in pyteal

        assert self.binary is not None
        assert len(self._template_values)
        assert len(args) == len(self._template_values)

        populate_program: list[Expr] = [
            (last_pos := ScratchVar(TealType.uint64)).store(Int(0)),
            (offset := ScratchVar(TealType.uint64)).store(Int(0)),
            (curr_val := ScratchVar(TealType.bytes)).store(Bytes("")),
            (buff := ScratchVar(TealType.bytes)).store(Bytes("")),
        ]

        for idx, tv in enumerate(self._template_values):
            # Add expressions to encode the values and insert them into the working buffer
            populate_program += [
                curr_val.store(Concat(encode_uvarint(Len(args[idx])), args[idx]))
                if tv.is_bytes
                else curr_val.store(encode_uvarint(args[idx])),
                buff.store(
                    Concat(
                        buff.load(),
                        Substring(
                            self.binary,
                            last_pos.load(),
                            Int(tv.pc),
                        ),
                        curr_val.load(),
                    )
                ),
                offset.store(offset.load() + Len(curr_val.load()) - Int(1)),
                last_pos.store(Int(tv.pc) + Int(1)),
            ]

        # append the bytes from the last template variable to the end
        populate_program += [
            buff.store(Concat(buff.load(), Suffix(self.binary, last_pos.load()))),
            buff.load(),
        ]

        @Subroutine(TealType.bytes)
        def populate_template_program():
            return Seq(*populate_program)

        return populate_template_program()

    def template_hash(self, *args) -> Expr:
        """
        returns an expression that will generate the expected
        hash given some set of values that should be included in the logic itself
        """
        return Sha512_256(
            Concat(Bytes(PROGRAM_DOMAIN_SEPARATOR), self.populate_template_expr(*args))
        )


class AppPrecompile:
    """
    AppPrecompile allows a smart contract to signal that some child Application
    should be fully compiled prior to constructing its own program.
    """

    def __init__(self, app: "Application"):
        #: The App to be used and compiled before it's parent
        self.app: "Application" = app
        #: The App's approval program as a Precompile
        self.approval: Precompile = Precompile("")
        #: The App's clear program as a Precompile
        self.clear: Precompile = Precompile("")

    def compile(self, client: AlgodClient):
        """fully compile this app precompile by recursively compiling children depth first

        Note:
            Must be called (even indirectly) prior to using the ``approval`` and ``clear`` fields
        """
        for p in self.app.precompiles.values():
            p.compile(client)

        # at this point, we should have all the dependant logic built
        # so we can compile the app teal
        approval, clear = self.app.compile(client)
        self.approval = Precompile(approval)
        self.clear = Precompile(clear)
        if self.approval._binary is None:
            self.approval.assemble(client)
        if self.clear._binary is None:
            self.clear.assemble(client)

    def get_create_config(self) -> dict[TxnField, Expr]:
        """get a dictionary of the fields and values that should be set when creating this application that can be passed directly to the InnerTxnBuilder.Execute method"""
        assert self.approval._binary is not None
        assert self.clear._binary is not None

        return {
            TxnField.type_enum: TxnType.ApplicationCall,
            TxnField.approval_program: self.approval.binary,
            TxnField.clear_state_program: self.clear.binary,
            TxnField.local_num_byte_slices: Int(self.app.acct_state.num_byte_slices),
            TxnField.local_num_uints: Int(self.app.acct_state.num_uints),
            TxnField.global_num_byte_slices: Int(self.app.app_state.num_byte_slices),
            TxnField.global_num_uints: Int(self.app.app_state.num_uints),
            TxnField.extra_program_pages: Int(
                num_extra_program_pages(self.approval._binary, self.clear._binary)
            ),
        }


class LSigPrecompile:
    """
    LSigPrecompile allows a smart contract to signal that some child Logic Signature
    should be fully compiled prior to constructing its own program.
    """

    def __init__(self, lsig: "LogicSignature"):
        #: the LogicSignature to be used and compiled before it's parent
        self.lsig: "LogicSignature" = lsig

        #: The LogicSignature's logic as a Precompile
        self.logic: Precompile = Precompile("")

    def compile(self, client: AlgodClient):
        """
        fully compile this lsig precompile by recursively compiling children depth first

        Note:
            Must be called (even indirectly) prior to using the ``approval`` and ``clear`` fields
        """
        for p in self.lsig.precompiles.values():
            p.compile(client)

        # at this point, we should have all the dependant logic built
        # so we can compile the lsig teal
        self.logic = Precompile(self.lsig.compile(client))

        if self.logic._binary is None:
            self.logic.assemble(client)

    def template_signer(self, *args) -> LogicSigTransactionSigner:
        """Get the Signer object for a populated version of the template contract"""
        return LogicSigTransactionSigner(
            LogicSigAccount(self.logic.populate_template(*args))
        )

    def signer(self) -> LogicSigTransactionSigner:
        """
        signer returns a LogicSigTransactionSigner to be used with
        an ApplicationClient or AtomicTransactionComposer.

        It should only be used for non templated Precompiles.
        """
        return LogicSigTransactionSigner(LogicSigAccount(self.logic._binary))


@dataclass
class ProgramAssertion:
    line: int
    message: str


def _gather_asserts(program: str, src_map: SourceMap) -> dict[int, ProgramAssertion]:
    asserts: dict[int, ProgramAssertion] = {}

    program_lines = program.split("\n")
    for idx, line in enumerate(program_lines):
        # Take only the first chunk before spaces
        line = line.split(" ").pop()
        if line != "assert":
            continue

        pc = src_map.get_pcs_for_line(idx)[0]

        # TODO: this will be wrong for multiline comments
        line_before = program_lines[idx - 1]
        if not line_before.startswith("//"):
            continue

        asserts[pc] = ProgramAssertion(idx, line_before.strip("// "))

    return asserts


def py_encode_uvarint(integer: int) -> bytes:
    """Encodes an integer as an uvarint.
    :param integer: the integer to encode
    :return: bytes containing the integer encoded as an uvarint
    """

    def to_byte(integer: int) -> int:
        return integer & 0b1111_1111

    buffer: bytearray = bytearray()

    while integer >= 0b1000_0000:
        buffer.append(to_byte(integer) | 0b1000_0000)
        integer >>= 7

    buffer.append(to_byte(integer))

    return bytes(buffer)



from typing import Final
from pyteal import (
    abi,
    InnerTxn,
    InnerTxnBuilder,
    Int,
    Seq,
    TealType,
    TxnField,
    TxnType,
)
from beaker import *
from beaker.precompile import AppPrecompile, LSigPrecompile


class LSig(LogicSignature):
    pass


class Child1(Application):
    counter: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        default=Int(0),
    )

    @create
    def create(self):
        return Seq(
            self.initialize_application_state(),
        )

    @external
    def increment_counter(self, *, output: abi.Uint64):
        """Increment the counter global state."""
        return Seq(
            self.counter.increment(),
            output.set(self.counter.get()),
        )


class Child2(Application):
    lsig: LSigPrecompile = LSigPrecompile(LSig())

    @external(read_only=True)
    def get_lsig_addr(self, *, output: abi.Address):
        return output.set(self.lsig.logic.hash())


class Parent(Application):
    child_1: AppPrecompile = AppPrecompile(Child1())
    child_2: AppPrecompile = AppPrecompile(Child2())

    @external
    def create_child_1(self, *, output: abi.Uint64):
        """Create a new child app."""
        return Seq(
            InnerTxnBuilder.Execute(
                {
                    TxnField.type_enum: TxnType.ApplicationCall,
                    TxnField.approval_program: self.child_1.approval.binary,
                    TxnField.clear_state_program: self.child_1.clear.binary,
                    TxnField.global_num_uints: Int(1),
                }
            ),
            output.set(InnerTxn.created_application_id()),
        )

    @external
    def create_child_2(self, *, output: abi.Uint64):
        """Create a new child app."""
        return Seq(
            InnerTxnBuilder.Execute(
                {
                    TxnField.type_enum: TxnType.ApplicationCall,
                    TxnField.approval_program: self.child_2.approval.binary,
                    TxnField.clear_state_program: self.child_2.clear.binary,
                    TxnField.global_num_uints: Int(1),
                }
            ),
            output.set(InnerTxn.created_application_id()),
        )


class Grandparent(Application):
    parent: AppPrecompile = AppPrecompile(Parent())

    @external
    def create_parent(self, *, output: abi.Uint64):
        """Create a new parent app."""
        return Seq(
            InnerTxnBuilder.Execute(
                {
                    TxnField.type_enum: TxnType.ApplicationCall,
                    TxnField.approval_program: self.parent.approval.binary,
                    TxnField.clear_state_program: self.parent.clear.binary,
                }
            ),
            output.set(InnerTxn.created_application_id()),
        )



import re
from copy import copy
from algosdk.source_map import SourceMap

LOGIC_ERROR = "TransactionPool.Remember: transaction ([A-Z0-9]+): logic eval error: (.*). Details: pc=([0-9]+), opcodes=.*"


def parse_logic_error(error_str: str) -> tuple[str, str, int]:
    matches = re.match(LOGIC_ERROR, error_str)
    if matches is None:
        return "", "", 0

    txid = matches.group(1)
    msg = matches.group(2)
    pc = int(matches.group(3))

    return txid, msg, pc


class LogicException(Exception):
    def __init__(
        self,
        logic_error: Exception,
        program: str,
        map: SourceMap,
    ):
        self.logic_error = logic_error
        self.logic_error_str = str(logic_error)

        self.program = program
        self.map = map
        self.lines = program.split("\n")

        self.txid, self.msg, self.pc = parse_logic_error(self.logic_error_str)
        self.line_no = self.map.get_line_for_pc(self.pc)

    def __str__(self):
        return f"Txn {self.txid} had error '{self.msg}' at PC {self.pc} and Source Line {self.line_no}: \n\n\t{self.trace()}"

    def trace(self, lines: int = 5) -> str:
        program_lines = copy(self.lines)
        program_lines[self.line_no] += "\t\t<-- Error"
        lines_before = max(0, self.line_no - lines)
        lines_after = min(len(program_lines), self.line_no + lines)
        return "\n\t".join(program_lines[lines_before:lines_after])



from typing import Final
from pyteal import *
from beaker import *

if __name__ == "__main__":
    from wormhole import ContractTransferVAA, WormholeTransfer
else:
    from .wormhole import ContractTransferVAA, WormholeTransfer


class OracleDataCache(WormholeTransfer):
    """
    Stores price feed in application state keyed by timestamp

    TODO: more than 64 vals lol
    """

    class OracleData(abi.NamedTuple):
        timestamp: abi.Field[abi.Uint64]
        price: abi.Field[abi.Uint64]
        confidence: abi.Field[abi.Uint64]

    prices: Final[ReservedApplicationStateValue] = ReservedApplicationStateValue(
        stack_type=TealType.bytes, max_keys=64
    )

    def handle_transfer(
        self, ctvaa: ContractTransferVAA, *, output: abi.DynamicBytes
    ) -> Expr:
        """
        invoked from parent class `portal_transfer` after parsing the VAA into
        abi vars
        """
        return Seq(
            # TODO: assert foreign sender? Should be in provided contract?
            # Do this once, since `get`` incurs a couple op penalty over `load`
            (payload := ScratchVar()).store(ctvaa.payload.get()),
            # Read vals from json
            (timestamp := abi.Uint64()).set(
                JsonRef.as_uint64(payload.load(), Bytes("ts"))
            ),
            (price := abi.Uint64()).set(
                JsonRef.as_uint64(payload.load(), Bytes("price"))
            ),
            (confidence := abi.Uint64()).set(
                JsonRef.as_uint64(payload.load(), Bytes("confidence"))
            ),
            # Construct named tuple for storage
            (od := abi.make(OracleDataCache.OracleData)).set(
                timestamp, price, confidence
            ),
            # Write to app state
            self.prices[timestamp].set(od.encode()),
            # echo the payload
            output.set(ctvaa.payload),
        )

    @external
    def lookup(self, ts: abi.Uint64, *, output: OracleData):
        return output.decode(self.prices[ts].get_must())


if __name__ == "__main__":
    OracleDataCache().dump("./spec")



from dataclasses import dataclass
import numpy as np
from copy import copy

# Python impl of constant product


@dataclass
class ConstantProductInvariant:
    a_supply: int
    b_supply: int

    max_pool_supply: int
    pool_supply: int

    scale: int
    fee: int

    def mint(self, a: int, b: int) -> int:
        a_rat = (a * self.scale) // self.a_supply
        b_rat = (b * self.scale) // self.b_supply

        # Use the min of the 2 scaled assets
        to_mint = (min(a_rat, b_rat) * self.issued()) // self.scale

        if to_mint == 0:
            return 0

        # compute first, then tweak balances
        self.a_supply += a
        self.b_supply += b
        self.pool_supply -= to_mint

        return to_mint

    def burn(self, amount) -> tuple[int, int]:
        burn_a = (self.a_supply * amount) // self.issued()
        burn_b = (self.b_supply * amount) // self.issued()

        if burn_a == 0 or burn_b == 0:
            return 0, 0

        # computed first, then tweak balances
        self.a_supply -= burn_a
        self.b_supply -= burn_b
        self.pool_supply += amount

        return burn_a, burn_b

    def swap(self, amount: int, is_a: bool) -> int:

        if is_a:
            swap_amt = self._get_tokens_to_swap(amount, self.a_supply, self.b_supply)
            if swap_amt == 0:
                return 0
            self.a_supply += amount
            self.b_supply -= swap_amt
            return swap_amt

        swap_amt = self._get_tokens_to_swap(amount, self.b_supply, self.a_supply)
        if swap_amt == 0:
            return 0

        self.b_supply += amount
        self.a_supply -= swap_amt
        return swap_amt

    def _get_tokens_to_swap(self, in_amount, in_supply, out_supply) -> int:
        assert in_supply > 0
        assert out_supply > 0
        """ Constant product swap method with fixed input

            X * Y = K

            X, Y are current supply of assets, goal is to keep K the same after adding in_amt and subtracting out amt 

            With no fees:
            ------------

            (X + X_in) * (Y - Y_out) = X*Y
            *Algebra happens*
            Y - ( X*Y / (X + X_in)) = Y_out

            With fees:
            ----------

            fee_factor = scale - fee (ex 1000, 3 for a .3% fee)

            X_adj = X_in * (fee_factor)

            (X + X_adj) * (Y - Y_out) = X*Y
            *Algebra happens*
            Y - ( X*Y / (X + X_adj)) = Y_out

            Or as in [Tinyman](https://github.com/tinymanorg/tinyman-contracts-v1/blob/main/contracts/validator_approval.teal#L1000): 
            Y_out = (X_adj * Y) / (X * scale + X_adj)
        """

        # Simple, no fees
        # out_amt = out_supply - ((in_supply * out_supply) / (in_supply + in_amount))

        # with fees
        factor = self.scale - self.fee
        out_amt = (in_amount * factor * out_supply) / (
            (in_supply * self.scale) + (in_amount * factor)
        )

        return int(out_amt)

    def scaled_ratio(self) -> int:
        return int(self.ratio() * self.scale)

    def issued(self) -> int:
        return self.max_pool_supply - self.pool_supply

    def ratio(self):
        return self.a_supply / self.b_supply


class Simulator:
    def __init__(self):
        self.cpi = ConstantProductInvariant(
            a_supply=int(3e7),
            b_supply=int(1e6),
            max_pool_supply=int(1e9),
            pool_supply=int(1e9) - 100000,
            scale=1000,
            fee=3,
        )

        self.states = []

    def run_mints(self, num: int = 100):
        self.sizes = np.random.randint(100, 1000, num)
        for size in self.sizes:
            # force them to be the right ratio
            a_size = size * self.cpi.ratio()
            b_size = size

            self.cpi.mint(a_size, b_size)
            self.states.append(copy(self.cpi))

    def run_burns(self, num: int = 100):
        self.sizes = np.random.randint(100, 10000000, num)
        for size in self.sizes:
            # Get a reasonable size given number of issued tokens
            # should be ~ 1% --> 0.0001% burns
            size = self.cpi.issued() // size
            self.cpi.burn(size)
            self.states.append(copy(self.cpi))

    def run_swaps(self, num: int = 100):

        self.sizes = np.random.randint(10, 1000, num)
        for idx, size in enumerate(self.sizes):
            a_swap = (idx + size) % 2 == 0
            # Re-size if its an a_swap
            size *= self.cpi.ratio() if a_swap else 1
            _ = self.cpi.swap(size, a_swap)
            self.states.append(copy(self.cpi))

    def run_mix(self, num=100):
        nums = np.random.randint(1, 4, num)
        for idx, runs in enumerate(nums):
            v = (idx + runs) % 3
            match v:
                case 0:
                    self.run_mints(runs)
                case 1:
                    self.run_burns(runs)
                case 2:
                    self.run_swaps(runs)

    def get_states_for(self, k: str) -> list[int]:
        if hasattr(self.cpi, k):
            return [getattr(c, k) for c in self.states]
        return []

    def get_states(self) -> dict[str, list[int]]:
        states: dict[str, list[int]] = {}
        for key in self.cpi.__annotations__.keys():
            states[key] = self.get_states_for(key)

        del states["scale"]
        del states["fee"]
        del states["max_pool_supply"]

        states["ratio"] = [s.ratio() for s in self.states]
        return states



import pyteal as pt
from ._list import List
from beaker.application import Application
from beaker.decorators import external

options = pt.CompileOptions(version=pt.MAX_TEAL_VERSION, mode=pt.Mode.Application)


def test_list():
    l = List(pt.abi.Uint64, 100, name="l")

    assert l._elements == 100
    assert l._element_size == 8
    assert l._box_size == 8 * 100
    assert l.value_type == pt.abi.Uint64TypeSpec()

    item = l[pt.Int(10)]
    with pt.TealComponent.Context.ignoreExprEquality():
        assert item.name.__teal__(options) == pt.Bytes("l").__teal__(options)
        assert item.element_size.__teal__(options) == pt.Int(8).__teal__(options)
        assert item.idx.__teal__(options) == pt.Int(10).__teal__(options)

    expected, _ = pt.BoxExtract(
        pt.Bytes("l"), pt.Int(8) * pt.Int(10), pt.Int(8)
    ).__teal__(options)
    actual, _ = item.get().__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    val = pt.abi.Uint64()
    expected, _ = pt.BoxReplace(
        pt.Bytes("l"), pt.Int(8) * pt.Int(10), val.encode()
    ).__teal__(options)
    actual, _ = item.set(val).__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_list_app():
    class T(Application):
        l = List(pt.abi.Uint64, 100)

        @external
        def get(self, idx: pt.abi.Uint16, *, output: pt.abi.Uint64):
            return self.l[idx.get()].store_into(output)

        @external
        def set(self, idx: pt.abi.Uint16, val: pt.abi.Uint64):
            return self.l[idx.get()].set(val)

    t = T()
    assert len(t.approval_program) > 0



from pyteal import *
from beaker import *


class ExternalExample(Application):
    @create
    def create(self, input: abi.String, *, output: abi.String):
        return output.decode(input.encode())


def demo():

    app_client = client.ApplicationClient(
        sandbox.get_algod_client(),
        ExternalExample(),
        signer=sandbox.get_accounts().pop().signer,
    )
    app_client.create(input="yo")
    # print(result.return_value)


if __name__ == "__main__":
    demo()



from hashlib import sha256
from algosdk.atomic_transaction_composer import (
    TransactionWithSigner,
)
from algosdk.future import transaction

from beaker.client import ApplicationClient
from beaker.consts import milli_algo
from beaker.sandbox import get_algod_client, get_accounts

if __name__ == "__main__":
    from contract import ExpensiveApp
else:
    from .contract import ExpensiveApp


client = get_algod_client()

acct = get_accounts().pop()


def demo():
    # Create an Application client containing both an algod client and my app
    sp = client.suggested_params()
    # we need to cover 255 inner transactions + ours
    sp.flat_fee = True
    sp.fee = 256 * milli_algo
    app_client = ApplicationClient(
        client, ExpensiveApp(), signer=acct.signer, suggested_params=sp
    )

    # Create the applicatiion on chain, set the app id for the app client
    app_id, app_addr, txid = app_client.create()
    print(f"Created App with id: {app_id} and address addr: {app_addr} in tx: {txid}")

    txn = TransactionWithSigner(
        txn=transaction.PaymentTxn(acct.address, sp, app_addr, int(1e6)),
        signer=acct.signer,
    )
    result = app_client.call(ExpensiveApp.opup_bootstrap, ptxn=txn)
    print(f"Created op up app: {result.return_value}")

    input = "stuff"
    iters = 10

    # You can compose calls if you pre-defined your ATC
    # atc = AtomicTransactionComposer()
    # app_client.add_method_call(atc, app.hash_it, input=input, iters=iters)
    # result = atc.execute(client, 4)

    result = app_client.call(ExpensiveApp.hash_it, input=input, iters=iters)
    result_hash = bytes(result.return_value)

    local_hash = input.encode()
    for _ in range(iters):
        local_hash = sha256(local_hash).digest()

    assert result_hash == local_hash, f"Expected {local_hash} got {result_hash}"

    print(f"Successfully hashed {input},  {iters} times to produce {result_hash.hex()}")


if __name__ == "__main__":
    demo()



from typing import Final

from algosdk.abi import ABIType
from pyteal import abi, TealType, Int, Seq

from beaker import (
    Application,
    ReservedAccountStateValue,
    opt_in,
    external,
    sandbox,
    client,
    identity_key_gen,
)


class Structer(Application):

    # Our custom Struct
    class Order(abi.NamedTuple):
        item: abi.Field[abi.String]
        quantity: abi.Field[abi.Uint16]

    orders: Final[ReservedAccountStateValue] = ReservedAccountStateValue(
        stack_type=TealType.bytes,
        max_keys=16,
        key_gen=identity_key_gen,
    )

    @opt_in
    def opt_in(self):
        return self.acct_state.initialize()

    @external
    def place_order(self, order_number: abi.Uint8, order: Order):
        return self.orders[order_number].set(order.encode())

    @external(read_only=True)
    def read_item(self, order_number: abi.Uint8, *, output: Order):
        return output.decode(self.orders[order_number])

    @external
    def increase_quantity(self, order_number: abi.Uint8, *, output: Order):
        return Seq(
            # Read the order from state
            (new_order := Structer.Order()).decode(self.orders[order_number]),
            # Select out in the quantity attribute, its a TupleElement type
            # so needs to be stored somewhere
            (quant := abi.Uint16()).set(new_order.quantity),
            # Add 1 to quantity
            quant.set(quant.get() + Int(1)),
            (item := abi.String()).set(new_order.item),
            # We've gotta set all of the fields at the same time, but we can
            # borrow the item we already know about
            new_order.set(item, quant),
            # Write the new order to state
            self.orders[order_number].set(new_order.encode()),
            # Write new order to caller
            output.decode(new_order.encode()),
        )


def demo():

    # Create a codec from the python sdk
    order_codec = ABIType.from_string(str(Structer.Order().type_spec()))

    acct = sandbox.get_accounts().pop()

    # Create an Application client containing both an algod client and my app
    app_client = client.ApplicationClient(
        sandbox.get_algod_client(), Structer(), signer=acct.signer
    )

    # Create the applicatiion on chain, set the app id for the app client
    app_id, app_addr, txid = app_client.create()
    print(f"Created App with id: {app_id} and address addr: {app_addr} in tx: {txid}")

    # Since we're using local state, opt in
    app_client.opt_in()

    # Passing in a dict as an argument that should take a tuple according to the type spec
    order_number = 12
    order = {"quantity": 8, "item": "cubes"}
    app_client.call(Structer.place_order, order_number=order_number, order=order)

    # Get the order from the state field
    state_key = order_number.to_bytes(1, "big")
    stored_order = app_client.get_account_state(raw=True)[state_key]
    state_decoded = order_codec.decode(stored_order)

    print(
        f"We can get the order we stored from local state of the sender: {state_decoded}"
    )

    # Or we could call the read-only method, passing the order number
    result = app_client.call(Structer.read_item, order_number=order_number)
    abi_decoded = order_codec.decode(result.raw_value)
    print(f"Decoded result: {abi_decoded}")

    # Update the order to increase the quantity
    result = app_client.call(Structer.increase_quantity, order_number=order_number)
    increased_decoded = order_codec.decode(result.raw_value)
    print(
        f"Let's add 1 to the struct, update state, and return the updated version: {increased_decoded}"
    )

    # And read it back out from state
    state_key = order_number.to_bytes(1, "big")
    stored_order = app_client.get_account_state(raw=True)[state_key]
    state_decoded = order_codec.decode(stored_order)
    print(f"And it's been updated: {state_decoded}")


if __name__ == "__main__":
    # import json
    # s = Structer()
    # with open("structer.json", "w") as f:
    #    f.write(json.dumps((s.application_spec())))

    demo()



from typing import Final

from beaker.client import ApplicationClient, LogicException
from beaker import sandbox

from pyteal import abi, TealType, Global, Int, Seq
from beaker.application import Application
from beaker.state import ApplicationStateValue
from beaker.decorators import external, create, Authorize


class CounterApp(Application):

    counter: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        descr="A counter for showing how to use application state",
    )

    @create
    def create(self):
        return self.initialize_application_state()

    @external(authorize=Authorize.only(Global.creator_address()))
    def increment(self, *, output: abi.Uint64):
        """increment the counter"""
        return Seq(
            self.counter.set(self.counter + Int(1)),
            output.set(self.counter),
        )

    @external(authorize=Authorize.only(Global.creator_address()))
    def decrement(self, *, output: abi.Uint64):
        """decrement the counter"""
        return Seq(
            self.counter.set(self.counter - Int(1)),
            output.set(self.counter),
        )


def demo():
    client = sandbox.get_algod_client()

    accts = sandbox.get_accounts()
    acct = accts.pop()

    # Create an Application client containing both an algod client and my app
    app_client = ApplicationClient(client, CounterApp(), signer=acct.signer)

    # Create the applicatiion on chain, set the app id for the app client
    app_id, app_addr, txid = app_client.create()
    print(f"Created App with id: {app_id} and address addr: {app_addr} in tx: {txid}")

    app_client.call(CounterApp.increment)
    app_client.call(CounterApp.increment)
    app_client.call(CounterApp.increment)
    result = app_client.call(CounterApp.increment)
    print(f"Currrent counter value: {result.return_value}")

    result = app_client.call(CounterApp.decrement)
    print(f"Currrent counter value: {result.return_value}")

    try:
        # Try to call the increment method with a different signer, it should fail
        # since we have the auth check
        other_acct = accts.pop()
        other_client = app_client.prepare(signer=other_acct.signer)
        other_client.call(CounterApp.increment)
    except LogicException as e:
        print("App call failed as expected.")
        print(e)


if __name__ == "__main__":
    ca = CounterApp()
    demo()



from pyteal import (
    abi,
    TealType,
    Expr,
    TealTypeError,
    Seq,
    Assert,
    BoxPut,
    BoxDelete,
    BoxGet,
    Pop,
    Concat,
)


class Mapping:
    """Mapping provides an abstraction to store some typed data in a box keyed with a typed key"""

    def __init__(
        self,
        key_type: type[abi.BaseType],
        value_type: type[abi.BaseType],
        prefix: Expr | None = None,
    ):
        """Initialize a Mapping object with details about storage

        Args:
            key_type: The type that will be used for the key. This type *MUST* encode to a byte string of < 64 bytes or it will fail at runtime.
            value_type:  The type to be stored in the box.
            prefix (Optional): An optional argument to prefix the key, providing a name space in order to avoid collisions with other mappings using the same keys
        """
        self._key_type = key_type
        self._key_type_spec = abi.type_spec_from_annotation(key_type)

        self._value_type = value_type
        self._value_type_spec = abi.type_spec_from_annotation(value_type)

        if isinstance(prefix, Expr) and prefix.type_of() != TealType.bytes:
            raise TealTypeError(prefix.type_of(), TealType.bytes)

        self.prefix = prefix

    def _prefix_key(self, key: Expr) -> Expr:
        if self.prefix is not None:
            return Concat(self.prefix, key)
        return key

    def __getitem__(self, key: abi.BaseType | Expr) -> "MapElement":
        match key:
            case abi.BaseType():
                if key.type_spec() != self._key_type_spec:
                    raise TealTypeError(key.type_spec(), self._key_type_spec)
                return MapElement(self._prefix_key(key.encode()), self._value_type)
            case Expr():
                if key.type_of() != TealType.bytes:
                    raise TealTypeError(key.type_of(), TealType.bytes)
                return MapElement(self._prefix_key(key), self._value_type)


class MapElement:
    """Container type for a specific box key and type"""

    def __init__(self, key: Expr, value_type: type[abi.BaseType]):
        assert key.type_of() == TealType.bytes, TealTypeError(
            key.type_of(), TealType.bytes
        )

        self.key = key
        self._value_type = value_type

    def exists(self) -> Expr:
        """check to see if a box with this key exists."""
        return Seq(maybe := BoxGet(self.key), maybe.hasValue())

    def store_into(self, val: abi.BaseType) -> Expr:
        """decode the bytes from this box into an abi type.

        Args:
            val: An instance of the type to be populated with the bytes from the box
        """
        return val.decode(self.get())

    def get(self) -> Expr:
        """get the bytes from this box."""
        return Seq(maybe := BoxGet(self.key), Assert(maybe.hasValue()), maybe.value())

    def set(self, val: abi.BaseType | Expr) -> Expr:
        """overwrites the contents of the box with the provided value.

        Args:
            val: An instance of the type or an Expr that evaluates to bytes
        """
        match val:
            case abi.BaseType():
                if not isinstance(val, self._value_type):
                    raise TealTypeError(val.__class__, self._value_type)
                return Seq(
                    # delete the old one
                    Pop(BoxDelete(self.key)),
                    # write the new one
                    BoxPut(self.key, val.encode()),
                )
            case Expr():
                if val.type_of() != TealType.bytes:
                    raise TealTypeError(val.type_of(), TealType.bytes)
                return BoxPut(self.key, val)

    def delete(self) -> Expr:
        """delete the box at this key"""
        return BoxDelete(self.key)



from beaker.client import ApplicationClient, LogicException
from beaker.sandbox import get_algod_client, get_accounts

# Leet hax, ignore
if __name__ == "__main__":
    from contract import StateExample
else:
    from .contract import StateExample


def demo():

    accts = get_accounts()

    acct = accts.pop()

    client = get_algod_client()

    app = StateExample()

    app_client = ApplicationClient(client, StateExample(), signer=acct.signer)
    app_id, app_address, transaction_id = app_client.create()
    print(
        f"DEPLOYED: App ID: {app_id} Address: {app_address} Transaction ID: {transaction_id}"
    )

    app_client.opt_in()
    print("Opted in")

    app_client.call(StateExample.set_account_state_val, v=123)
    app_client.call(StateExample.incr_account_state_val, v=1)
    result = app_client.call(StateExample.get_account_state_val)
    print(f"Set/get acct state result: {result.return_value}")

    app_client.call(StateExample.set_dynamic_account_state_val, k=123, v="stuff")
    result = app_client.call(StateExample.get_dynamic_account_state_val, k=123)
    print(f"Set/get dynamic acct state result: {result.return_value}")

    try:
        app_client.call(StateExample.set_app_state_val, v="Expect fail")
    except LogicException as e:
        print(f"Task failed successfully: {e}")
        result = app_client.call(StateExample.get_app_state_val)
        print(f"Set/get app state result: {result.return_value}")

    app_client.call(StateExample.set_dynamic_app_state_val, k=15, v=123)
    result = app_client.call(StateExample.get_dynamic_app_state_val, k=15)
    print(f"Set/get dynamic app state result: {result.return_value}")

    msg = "write this message please and make it readable"

    # Account state blob
    app_client.call(StateExample.write_acct_blob, v=msg)
    result = app_client.call(StateExample.read_acct_blob)
    got_msg = bytes(result.return_value[: len(msg)]).decode()
    assert msg == got_msg
    print(f"wrote and read the message to account state {got_msg}")

    # App state blob
    app_client.call(StateExample.write_app_blob, v=msg)
    # result = app_client.call(StateExample.read_app_blob)
    # got_msg = bytes(result.return_value[: len(msg)]).decode()
    # assert msg == got_msg
    print(f"wrote and read the message to application state {got_msg}")


if __name__ == "__main__":
    demo()



"""Module containing helper functions for testing PyTeal Utils."""
from typing import Any
from algosdk.atomic_transaction_composer import AtomicTransactionComposer

import pyteal as pt
from beaker import client, sandbox
from beaker import Application, external, delete, update, opt_in, close_out

algod_client = None
sandbox_accounts = None


def returned_int_as_bytes(i: int, bits: int = 64):
    return list(i.to_bytes(bits // 8, "big"))


class UnitTestingApp(Application):

    """Base unit testable application.

    There are 2 ways to use this class


    1) Initialize with a single Expr that returns bytes
        The bytes output from the Expr are returned from the abi method ``unit_test()[]byte``

    2) Subclass UnitTestingApp and override `unit_test`
        Any inputs or output may be specified but you're responsible for encoding the incoming
        arguments as a dict with keys matching the argument names of the custom `unit_test` method


    An instance of this class is passed to assert_output to check the return value against what you expect.
    """

    def __init__(self, expr_to_test: pt.Expr = None):
        self.expr = expr_to_test
        super().__init__()

    @delete
    def delete(self):
        return pt.Approve()

    @update
    def update(self):
        return pt.Approve()

    @opt_in
    def opt_in(self):
        return self.acct_state.initialize()

    @close_out
    def close_out(self):
        return pt.Approve()

    @external
    def opup(self):
        return pt.Approve()

    @external
    def unit_test(self, *, output: pt.abi.DynamicArray[pt.abi.Byte]):
        if self.expr is None:
            raise Exception(
                "Expression undefined. Either set the expr to test on init or override unit_test method"
            )
        return pt.Seq((s := pt.abi.String()).set(self.expr), output.decode(s.encode()))


def assert_output(
    app: UnitTestingApp,
    inputs: list[dict[str, Any]],
    outputs: list[Any],
    opups: int = 0,
):
    """
    Creates and calls the UnitTestingApp passed and compares the return value with the expected output

    :param app: An instance of a UnitTestingApp to make call against its `unit_test` method
    :param inputs: A list of dicts where each entry contains keys matching the input args for the `unit_test` method  and values corresponding to the type expected by the method
    :param outputs: A list of outputs to compare against the return value of the output of the `unit_test` method
    :param opups: A number of additional app call transactions to make to increase our budget

    """
    # TODO: make these avail in a pytest session context? pass them in directly?
    global algod_client, sandbox_accounts
    if algod_client is None:
        algod_client = sandbox.get_algod_client()

    if sandbox_accounts is None:
        sandbox_accounts = sandbox.get_accounts()

    app_client = client.ApplicationClient(
        algod_client, app, signer=sandbox_accounts[0].signer
    )
    app_client.create()

    has_state = app.acct_state.num_byte_slices + app.acct_state.num_uints > 0

    if has_state:
        app_client.opt_in()

    try:
        for idx, output in enumerate(outputs):
            input = {} if len(inputs) == 0 else inputs[idx]

            if opups > 0:
                atc = AtomicTransactionComposer()

                app_client.add_method_call(atc, app.unit_test, **input)
                for x in range(opups):
                    app_client.add_method_call(atc, app.opup, note=str(x).encode())

                try:
                    results = atc.execute(algod_client, 2)
                except Exception as e:
                    raise app_client.wrap_approval_exception(e)

                assert results.abi_results[0].return_value == output
            else:
                result = app_client.call(app.unit_test, **input)
                assert result.return_value == output
    except Exception as e:
        raise e
    finally:
        if has_state:
            app_client.close_out()
        app_client.delete()



from distutils.core import setup
from setuptools import find_packages


setup(
    name="beaker",
    version="0.4.0",
    python_requires=">=3.10",
    packages=find_packages(),
    license="MIT",
    long_description=open("README.md").read(),
    package_data={"beaker": ["py.typed"]},
    install_requires=["pyteal>=0.20.1", "py-algorand-sdk>=1.16.1"],
)



import pytest
from typing import Final, cast
from dataclasses import asdict
from Cryptodome.Hash import SHA512
import pyteal as pt
from pyteal.ast.abi import PaymentTransaction, AssetTransferTransaction

from beaker.state import (
    ReservedApplicationStateValue,
    ApplicationStateValue,
    AccountStateValue,
    ReservedAccountStateValue,
)

from beaker.errors import BareOverwriteError
from beaker.application import (
    Application,
    get_method_selector,
    get_method_signature,
    get_method_spec,
)
from beaker.decorators import (
    DefaultArgumentClass,
    external,
    get_handler_config,
    internal,
    create,
    opt_in,
    clear_state,
    close_out,
    update,
    delete,
)

options = pt.CompileOptions(mode=pt.Mode.Application, version=pt.MAX_TEAL_VERSION)


def test_empty_application():
    class EmptyApp(Application):
        pass

    ea = EmptyApp()

    assert ea.router.name == "EmptyApp", "Expected router name to match class"
    assert (
        ea.acct_state.num_uints
        + ea.acct_state.num_byte_slices
        + ea.app_state.num_uints
        + ea.app_state.num_byte_slices
        == 0
    ), "Expected no schema"

    assert len(ea.bare_externals.keys()) == len(
        EXPECTED_BARE_HANDLERS
    ), f"Expected {len(EXPECTED_BARE_HANDLERS)} bare handlers: {EXPECTED_BARE_HANDLERS}"
    assert (
        len(ea.approval_program) > 0
    ), "Expected approval program to be compiled to teal"
    assert len(ea.clear_program) > 0, "Expected clear program to be compiled to teal"
    assert len(ea.contract.methods) == 0, "Expected no methods in the contract"


def test_teal_version():
    class EmptyApp(Application):
        pass

    ea = EmptyApp(version=4)

    assert ea.teal_version == 4, "Expected teal v4"
    assert ea.approval_program.split("\n")[0] == "#pragma version 4"


def test_single_external():
    class Singleexternal(Application):
        @external
        def handle():
            return pt.Assert(pt.Int(1))

    sh = Singleexternal()

    assert len(sh.methods) == 1, "Expected a single external"
    assert sh.contract.get_method_by_name("handle") == get_method_spec(
        sh.handle
    ), "Expected contract method to match method spec"

    with pytest.raises(Exception):
        sh.contract.get_method_by_name("made up")


def test_method_override():
    class MethodOverride(Application):
        @external(name="handle")
        def handle_algo(self, txn: PaymentTransaction):
            return pt.Approve()

        @external(name="handle")
        def handle_asa(self, txn: AssetTransferTransaction):
            return pt.Approve()

    mo = MethodOverride()

    assert len(mo.methods) == 2, "Expected two externals"
    assert mo.methods["handle_algo"]
    assert mo.methods["handle_asa"]

    overlapping_methods = [
        method for method in mo.contract.methods if method.name == "handle"
    ]
    assert len(overlapping_methods) == 2
    assert get_method_spec(mo.handle_algo) in overlapping_methods
    assert get_method_spec(mo.handle_asa) in overlapping_methods


def test_bare():
    class Bare(Application):
        @create
        def create():
            return pt.Approve()

        @update
        def update():
            return pt.Approve()

        @delete
        def delete():
            return pt.Approve()

    bh = Bare()

    assert (
        len(bh.bare_externals) == 3
    ), "Expected 3 bare externals: create,update,delete"

    class FailBare(Application):
        @create
        def wrong_name():
            return pt.Approve()

    with pytest.raises(BareOverwriteError):
        bh = FailBare()


def test_mixed_bares():
    class MixedBare(Application):
        @create
        def create(self):
            return pt.Approve()

        @opt_in
        def opt_in(self, s: pt.abi.String):
            return pt.Assert(pt.Len(s.get()))

    mb = MixedBare()
    assert len(mb.bare_externals) == 1
    assert len(mb.methods) == 1


def test_bare_external():
    class BareExternal(Application):
        @create
        def create(self, s: pt.abi.String):
            return pt.Assert(pt.Len(s.get()))

        @opt_in
        def opt_in(self, s: pt.abi.String):
            return pt.Assert(pt.Len(s.get()))

        @close_out
        def close_out(self, s: pt.abi.String):
            return pt.Assert(pt.Len(s.get()))

        @clear_state
        def clear_state(self, s: pt.abi.String):
            return pt.Assert(pt.Len(s.get()))

        @update
        def update(self, s: pt.abi.String):
            return pt.Assert(pt.Len(s.get()))

        @delete
        def delete(self, s: pt.abi.String):
            return pt.Assert(pt.Len(s.get()))

    be = BareExternal()
    assert len(be.bare_externals) == 0, "Should have no bare externals"
    assert (
        len(be.contract.methods) == 6
    ), "should have create, optin, closeout, clearstate, update, delete"

    hc = get_handler_config(BareExternal.create)
    assert hc.method_config is not None
    confs = asdict(hc.method_config)
    assert confs["no_op"] == pt.CallConfig.CREATE
    del confs["no_op"]
    assert all([c == pt.CallConfig.NEVER for c in confs.values()])

    hc = get_handler_config(BareExternal.opt_in)
    assert hc.method_config is not None
    confs = asdict(hc.method_config)
    assert confs["opt_in"] == pt.CallConfig.CALL
    del confs["opt_in"]
    assert all([c == pt.CallConfig.NEVER for c in confs.values()])

    hc = get_handler_config(BareExternal.close_out)
    assert hc.method_config is not None
    confs = asdict(hc.method_config)
    assert confs["close_out"] == pt.CallConfig.CALL
    del confs["close_out"]
    assert all([c == pt.CallConfig.NEVER for c in confs.values()])

    hc = get_handler_config(BareExternal.clear_state)
    assert hc.method_config is not None
    confs = asdict(hc.method_config)
    assert confs["clear_state"] == pt.CallConfig.CALL
    del confs["clear_state"]
    assert all([c == pt.CallConfig.NEVER for c in confs.values()])

    hc = get_handler_config(BareExternal.update)
    assert hc.method_config is not None
    confs = asdict(hc.method_config)
    assert confs["update_application"] == pt.CallConfig.CALL
    del confs["update_application"]
    assert all([c == pt.CallConfig.NEVER for c in confs.values()])

    hc = get_handler_config(BareExternal.delete)
    assert hc.method_config is not None
    confs = asdict(hc.method_config)
    assert confs["delete_application"] == pt.CallConfig.CALL
    del confs["delete_application"]
    assert all([c == pt.CallConfig.NEVER for c in confs.values()])


def test_subclass_application():
    class SuperClass(Application):
        @external
        def handle():
            return pt.Assert(pt.Int(1))

    class SubClass(SuperClass):
        pass

    sc = SubClass()
    assert len(sc.methods) == 1, "Expected single method"
    assert sc.contract.get_method_by_name("handle") == get_method_spec(
        sc.handle
    ), "Expected contract method to match method spec"

    class OverrideSubClass(SuperClass):
        @external
        def handle():
            return pt.Assert(pt.Int(2))

    osc = OverrideSubClass()
    assert len(osc.methods) == 1, "Expected single method"
    assert osc.contract.get_method_by_name("handle") == get_method_spec(
        osc.handle
    ), "Expected contract method to match method spec"


def test_app_state():
    class BasicAppState(Application):
        uint_val: Final[ApplicationStateValue] = ApplicationStateValue(
            stack_type=pt.TealType.uint64
        )
        byte_val: Final[ApplicationStateValue] = ApplicationStateValue(
            stack_type=pt.TealType.bytes
        )

    app = BasicAppState()

    assert app.app_state.num_uints == 1, "Expected 1 int"
    assert app.app_state.num_byte_slices == 1, "Expected 1 byte slice"

    class ReservedAppState(BasicAppState):
        uint_dynamic: Final[
            ReservedApplicationStateValue
        ] = ReservedApplicationStateValue(stack_type=pt.TealType.uint64, max_keys=10)
        byte_dynamic: Final[
            ReservedApplicationStateValue
        ] = ReservedApplicationStateValue(stack_type=pt.TealType.bytes, max_keys=10)

    app = ReservedAppState()
    assert app.app_state.num_uints == 11, "Expected 11 ints"
    assert app.app_state.num_byte_slices == 11, "Expected 11 byte slices"


def test_acct_state():
    class BasicAcctState(Application):
        uint_val: Final[AccountStateValue] = AccountStateValue(
            stack_type=pt.TealType.uint64
        )
        byte_val: Final[AccountStateValue] = AccountStateValue(
            stack_type=pt.TealType.bytes
        )

    app = BasicAcctState()

    assert app.acct_state.num_uints == 1, "Expected 1 int"
    assert app.acct_state.num_byte_slices == 1, "Expected 1 byte slice"

    class ReservedAcctState(BasicAcctState):
        uint_dynamic: Final[ReservedAccountStateValue] = ReservedAccountStateValue(
            stack_type=pt.TealType.uint64, max_keys=5
        )
        byte_dynamic: Final[ReservedAccountStateValue] = ReservedAccountStateValue(
            stack_type=pt.TealType.bytes, max_keys=5
        )

    app = ReservedAcctState()
    assert app.acct_state.num_uints == 6, "Expected 6 ints"
    assert app.acct_state.num_byte_slices == 6, "Expected 6 byte slices"


def test_internal():
    from beaker.decorators import internal

    class Internal(Application):
        @create
        def create(self):
            return pt.Seq(
                pt.Pop(self.internal_meth()),
                pt.Pop(self.internal_meth_no_self()),
                pt.Pop(self.subr_no_self()),
            )

        @external(method_config=pt.MethodConfig(no_op=pt.CallConfig.CALL))
        def otherthing():
            return pt.Seq(
                pt.Pop(Internal.internal_meth_no_self()),
                pt.Pop(Internal.subr_no_self()),
            )

        @internal(pt.TealType.uint64)
        def internal_meth(self):
            return pt.Int(1)

        @internal(pt.TealType.uint64)
        def internal_meth_no_self():
            return pt.Int(1)

        # Cannot be called with `self` specified
        @pt.Subroutine(pt.TealType.uint64)
        def subr(self):
            return pt.Int(1)

        @pt.Subroutine(pt.TealType.uint64)
        def subr_no_self():
            return pt.Int(1)

    i = Internal()
    assert len(i.methods) == 1, "Expected 1 ABI method"

    # Test with self
    meth = cast(pt.SubroutineFnWrapper, i.internal_meth)
    expected = pt.SubroutineCall(meth.subroutine, []).__teal__(options)

    actual = i.internal_meth().__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    # Cannot call it without instantiated object
    with pytest.raises(Exception):
        Internal.internal_meth()

    # Test with no self
    meth = cast(pt.SubroutineFnWrapper, i.internal_meth_no_self)
    expected = pt.SubroutineCall(meth.subroutine, []).__teal__(options)

    actual = i.internal_meth_no_self().__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    # Cannot call a subroutine that references self
    with pytest.raises(pt.TealInputError):
        i.subr()

    # Test subr with no self
    meth = cast(pt.SubroutineFnWrapper, i.subr_no_self)
    expected = pt.SubroutineCall(meth.subroutine, []).__teal__(options)
    actual = i.subr_no_self().__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_default_param_state():
    class Hinty(Application):
        asset_id = ApplicationStateValue(pt.TealType.uint64, default=pt.Int(123))

        @external
        def hintymeth(self, num: pt.abi.Uint64, aid: pt.abi.Asset = asset_id):
            return pt.Assert(aid.asset_id() == self.asset_id)

    h = Hinty()

    assert h.hintymeth.__name__ in h.hints, "Expected a hint available for the method"

    hint = h.hints[h.hintymeth.__name__]

    assert "aid" in hint.default_arguments, "Expected annotation available for param"

    default = hint.default_arguments["aid"]

    assert default.resolvable_class == DefaultArgumentClass.GlobalState
    assert (
        default.resolve_hint() == Hinty.asset_id.str_key()
    ), "Expected the hint to match the method spec"


def test_default_param_const():
    const_val = 123

    class Hinty(Application):
        @external
        def hintymeth(self, num: pt.abi.Uint64, aid: pt.abi.Asset = const_val):
            return pt.Assert(aid.asset_id() == pt.Int(const_val))

    h = Hinty()

    assert h.hintymeth.__name__ in h.hints, "Expected a hint available for the method"

    hint = h.hints[h.hintymeth.__name__]

    assert "aid" in hint.default_arguments, "Expected annotation available for param"

    default = hint.default_arguments["aid"]

    assert default.resolvable_class == DefaultArgumentClass.Constant
    assert (
        default.resolve_hint() == const_val
    ), "Expected the hint to match the method spec"


def test_default_read_only_method():
    const_val = 123

    class Hinty(Application):
        @external(read_only=True)
        def get_asset_id(self, *, output: pt.abi.Uint64):
            return output.set(pt.Int(const_val))

        @external
        def hintymeth(self, num: pt.abi.Uint64, aid: pt.abi.Asset = get_asset_id):
            return pt.Assert(aid.asset_id() == pt.Int(const_val))

    h = Hinty()

    assert h.hintymeth.__name__ in h.hints, "Expected a hint available for the method"

    hint = h.hints[h.hintymeth.__name__]

    assert "aid" in hint.default_arguments, "Expected annotation available for param"

    default = hint.default_arguments["aid"]

    assert default.resolvable_class == DefaultArgumentClass.ABIMethod
    assert (
        default.resolve_hint() == get_method_spec(Hinty.get_asset_id).dictify()
    ), "Expected the hint to match the method spec"


def test_app_spec():
    class Specd(Application):
        decl_app_val = ApplicationStateValue(pt.TealType.uint64)
        decl_acct_val = AccountStateValue(pt.TealType.uint64)

        @external(read_only=True)
        def get_asset_id(self, *, output: pt.abi.Uint64):
            return output.set(pt.Int(123))

        @external
        def annotated_meth(self, aid: pt.abi.Asset = get_asset_id):
            return pt.Assert(pt.Int(1))

        class Thing(pt.abi.NamedTuple):
            a: pt.abi.Field[pt.abi.Uint64]
            b: pt.abi.Field[pt.abi.Uint32]

        @external
        def struct_meth(self, thing: Thing):
            return pt.Approve()

    s = Specd()

    actual_spec = s.application_spec()

    get_asset_id_hints = {"read_only": True}
    annotated_meth_hints = {
        "default_arguments": {
            "aid": {
                "source": "abi-method",
                "data": {
                    "name": "get_asset_id",
                    "args": [],
                    "returns": {"type": "uint64"},
                },
            },
        }
    }
    struct_meth_hints = {
        "structs": {
            "thing": {"name": "Thing", "elements": [("a", "uint64"), ("b", "uint32")]}
        }
    }

    expected_hints = {
        "get_asset_id": get_asset_id_hints,
        "annotated_meth": annotated_meth_hints,
        "struct_meth": struct_meth_hints,
    }

    expected_schema = {
        "local": {
            "declared": {
                "decl_acct_val": {
                    "type": "uint64",
                    "key": "decl_acct_val",
                    "descr": "",
                }
            },
            "reserved": {},
        },
        "global": {
            "declared": {
                "decl_app_val": {
                    "type": "uint64",
                    "key": "decl_app_val",
                    "descr": "",
                }
            },
            "reserved": {},
        },
    }

    def dict_match(a: dict, e: dict) -> bool:
        for k, v in a.items():
            if type(v) is dict:
                if not dict_match(v, e[k]):
                    print(f"comparing {k} {v} {e[k]}")
                    return False
            else:
                if v != e[k]:
                    print(f"comparing {k}")
                    return False

        return True

    assert dict_match(actual_spec["hints"], expected_hints)
    assert dict_match(actual_spec["schema"], expected_schema)


EXPECTED_BARE_HANDLERS = [
    "create",
]


def test_struct_args():
    from algosdk.abi import Method, Argument, Returns

    class Structed(Application):
        class UserRecord(pt.abi.NamedTuple):
            addr: pt.abi.Field[pt.abi.Address]
            balance: pt.abi.Field[pt.abi.Uint64]
            nickname: pt.abi.Field[pt.abi.String]

        @external
        def structy(self, user_record: UserRecord):
            return pt.Assert(pt.Int(1))

    m = Structed()

    arg = Argument("(address,uint64,string)", name="user_record")
    ret = Returns("void")
    assert Method("structy", [arg], ret) == get_method_spec(m.structy)

    assert m.hints["structy"].structs == {
        "user_record": {
            "name": "UserRecord",
            "elements": [
                ("addr", "address"),
                ("balance", "uint64"),
                ("nickname", "string"),
            ],
        }
    }


def test_instance_vars():
    class Inst(Application):
        def __init__(self, v: str):
            self.v = pt.Bytes(v)
            super().__init__()

        @external
        def use_it(self):
            return pt.Log(self.v)

        @external
        def call_it(self):
            return self.use_it_internal()

        @internal(pt.TealType.none)
        def use_it_internal(self):
            return pt.Log(self.v)

    i1 = Inst("first")
    i2 = Inst("second")

    assert i1.approval_program.index("first") > 0, "Expected to see the string `first`"
    assert (
        i2.approval_program.index("second") > 0
    ), "Expected to see the string `second`"

    with pytest.raises(ValueError):
        i1.approval_program.index("second")

    with pytest.raises(ValueError):
        i2.approval_program.index("first")


def hashy(sig: str):
    chksum = SHA512.new(truncate="256")
    chksum.update(sig.encode())
    return chksum.digest()[:4]


def test_abi_method_details():
    @external
    def meth():
        return pt.Assert(pt.Int(1))

    expected_sig = "meth()void"
    expected_selector = hashy(expected_sig)

    assert get_method_signature(meth) == expected_sig
    assert get_method_selector(meth) == expected_selector

    def meth2():
        pass

    with pytest.raises(Exception):
        get_method_spec(meth2)

    with pytest.raises(Exception):
        get_method_signature(meth2)

    with pytest.raises(Exception):
        get_method_selector(meth2)



import pytest
import pyteal as pt
from algosdk import encoding

from beaker.application import Application
from beaker.decorators import external
from beaker.client import ApplicationClient
from beaker.sandbox import get_accounts, get_algod_client
from beaker.logic_signature import LogicSignature, TemplateVariable
from beaker.precompile import (
    AppPrecompile,
    LSigPrecompile,
    py_encode_uvarint,
)


def test_precompile_basic():
    class App(Application):
        class Lsig(LogicSignature):
            def evaluate(self):
                return pt.Seq(pt.Assert(pt.Int(1)), pt.Int(1))

        pc = LSigPrecompile(Lsig(version=6))

        @external
        def check_it(self):
            return pt.Assert(pt.Txn.sender() == self.pc.logic.hash())

    app = App()
    ac = ApplicationClient(get_algod_client(), app, signer=get_accounts().pop().signer)

    assert app.approval_program is None
    assert app.clear_program is None
    assert app.pc.logic._program_hash is None

    ac.build()

    assert app.approval_program is not None
    assert app.clear_program is not None
    assert app.pc.logic._program_hash is not None


TMPL_BYTE_VALS = [
    ("abc"),
    ("asdfasdfasdf"),
    (bytes(100)),
    ("的的的的的的的"),
]


@pytest.mark.parametrize("tmpl_val", TMPL_BYTE_VALS)
def test_templated_bytes(tmpl_val: str):
    class Lsig(LogicSignature):
        tv = TemplateVariable(pt.TealType.bytes)

        def evaluate(self):
            return pt.Seq(pt.Assert(pt.Len(self.tv)), pt.Int(1))

    class App(Application):
        pc = LSigPrecompile(Lsig(version=6))

        @external
        def check_it(self):
            return pt.Assert(
                pt.Txn.sender() == self.pc.logic.template_hash(pt.Bytes(tmpl_val))
            )

    app = App()
    ac = ApplicationClient(get_algod_client(), app, signer=get_accounts().pop().signer)

    assert app.approval_program is None
    assert app.clear_program is None
    assert app.pc.logic._program_hash is None

    ac.build()

    assert app.approval_program is not None
    assert app.clear_program is not None
    assert app.pc.logic._program_hash is not None

    populated_teal = app.pc.logic.populate_template(tmpl_val)

    vlen = len(tmpl_val)
    if type(tmpl_val) is str:
        vlen = len(tmpl_val.encode("utf-8"))

    assert len(populated_teal) == len(app.pc.logic._binary) + vlen + (
        len(py_encode_uvarint(vlen)) - 1
    )


TMPL_INT_VALS = [(10), (1000), (int(2.9e9))]


@pytest.mark.parametrize("tmpl_val", TMPL_INT_VALS)
def test_templated_ints(tmpl_val: int):
    class Lsig(LogicSignature):
        tv = TemplateVariable(pt.TealType.uint64)

        def evaluate(self):
            return pt.Seq(pt.Assert(self.tv), pt.Int(1))

    class App(Application):
        pc = LSigPrecompile(Lsig(version=6))

        @external
        def check_it(self):
            return pt.Assert(
                pt.Txn.sender() == self.pc.logic.template_hash(pt.Int(tmpl_val))
            )

    app = App()
    ac = ApplicationClient(get_algod_client(), app, signer=get_accounts().pop().signer)

    assert app.approval_program is None
    assert app.clear_program is None
    assert app.pc.logic._program_hash is None

    ac.build()

    assert app.approval_program is not None
    assert app.clear_program is not None
    assert app.pc.logic._program_hash is not None

    populated_teal = app.pc.logic.populate_template(tmpl_val)

    assert len(populated_teal) == len(app.pc.logic._binary) + (
        len(py_encode_uvarint(tmpl_val)) - 1
    )


class InnerLsig(LogicSignature):
    nonce = TemplateVariable(pt.TealType.bytes)

    def evaluate(self):
        return pt.Approve()


class InnerApp(Application):
    pass


class OuterApp(Application):
    child: AppPrecompile = AppPrecompile(InnerApp())
    lsig: LSigPrecompile = LSigPrecompile(InnerLsig())

    @external
    def doit(self, nonce: pt.abi.DynamicBytes, *, output: pt.abi.Uint64):
        return pt.Seq(
            pt.Assert(pt.Txn.sender() == self.lsig.logic.template_hash(nonce.get())),
            pt.InnerTxnBuilder.Execute(
                {
                    pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
                    pt.TxnField.approval_program: self.child.approval.binary,
                    pt.TxnField.clear_state_program: self.child.clear.binary,
                    pt.TxnField.fee: pt.Int(0),
                }
            ),
            output.set(pt.InnerTxn.created_application_id()),
        )


def test_nested_precompile():
    oa = OuterApp()

    # Nothing is available until we build out the app and all its precompiles
    assert oa.child.approval._program is not None
    assert oa.child.clear._program is not None
    assert oa.lsig.logic._program is not None

    assert len(oa.lsig.logic._template_values) == 0

    assert oa.child.approval._binary is None
    assert oa.child.clear._binary is None
    assert oa.lsig.logic._binary is None

    ac = ApplicationClient(
        client=get_algod_client(), app=oa, signer=get_accounts().pop().signer
    )
    ac.build()

    assert oa.child.approval._binary is not None
    assert oa.child.clear._binary is not None
    assert oa.lsig.logic._binary is not None

    assert len(oa.lsig.logic._template_values) == 1


def test_build_recursive():
    app = OuterApp()
    pc = AppPrecompile(app)
    pc.compile(get_algod_client())
    _check_app_precompiles(pc)


class LargeApp(Application):
    longBytes = 4092 * b"A"
    longBytes2 = 2048 * b"A"

    @external
    def compare_big_byte_strings(self):
        return pt.Assert(pt.Bytes(self.longBytes) != pt.Bytes(self.longBytes2))


def test_page_hash():
    class SmallApp(Application):
        pass

    small_precompile = AppPrecompile(SmallApp())
    small_precompile.compile(get_algod_client())
    _check_app_precompiles(small_precompile)


def test_extra_page_population():

    app = LargeApp()
    app_precompile = AppPrecompile(app)
    app_precompile.compile(get_algod_client())
    _check_app_precompiles(app_precompile)

    assert app_precompile.approval.program_pages is not None
    assert app_precompile.clear.program_pages is not None
    recovered_approval_binary = b""
    for approval_page in app_precompile.approval.program_pages:
        assert len(approval_page._hash_digest) == 32
        recovered_approval_binary += approval_page._binary

    recovered_clear_binary = b""
    for clear_page in app_precompile.clear.program_pages:
        recovered_clear_binary += clear_page._binary

    assert recovered_approval_binary == app_precompile.approval._binary
    assert recovered_clear_binary == app_precompile.clear._binary


def _check_app_precompiles(app_precompile: AppPrecompile):
    for _, p in app_precompile.app.precompiles.items():
        match p:
            case LSigPrecompile():
                _check_lsig_precompiles(p)
            case AppPrecompile():
                _check_app_precompiles(p)

    assert app_precompile.approval._program != ""
    assert app_precompile.approval._binary is not None
    assert app_precompile.approval.binary.byte_str != b""
    assert app_precompile.approval._map is not None
    assert app_precompile.approval._program_hash is not None
    assert app_precompile.approval._template_values == []

    assert len(app_precompile.approval.program_pages) > 0
    if len(app_precompile.approval.program_pages) == 1:
        assert app_precompile.approval.program_pages[
            0
        ]._hash_digest == encoding.decode_address(app_precompile.approval._program_hash)

    assert app_precompile.clear._program != ""
    assert app_precompile.clear._binary is not None
    assert app_precompile.clear.binary.byte_str != b""
    assert app_precompile.clear._map is not None
    assert app_precompile.clear._program_hash is not None
    assert app_precompile.clear._template_values == []
    assert len(app_precompile.clear.program_pages) > 0
    if len(app_precompile.clear.program_pages) == 1:
        assert app_precompile.clear.program_pages[
            0
        ]._hash_digest == encoding.decode_address(app_precompile.clear._program_hash)


def _check_lsig_precompiles(lsig_precompile: LSigPrecompile):
    for _, p in lsig_precompile.lsig.precompiles.items():
        match p:
            case LSigPrecompile():
                _check_lsig_precompiles(p)
            case AppPrecompile():
                _check_app_precompiles(p)

    assert lsig_precompile.logic._program != ""
    assert lsig_precompile.logic._binary is not None
    assert lsig_precompile.logic.binary.byte_str != b""
    assert lsig_precompile.logic._map is not None
    assert lsig_precompile.logic._program_hash is not None
    assert len(lsig_precompile.logic._template_values) == len(
        lsig_precompile.lsig.template_variables
    )



import math as pymath

import pyteal as pt
import beaker as bkr

from beaker.testing.unit_testing_helpers import UnitTestingApp, assert_output

import beaker.lib.math as math


def test_even():
    class EvenTester(UnitTestingApp):
        @bkr.external
        def unit_test(self, num: pt.abi.Uint64, *, output: pt.abi.Bool):
            return output.set(math.even(num.get()))

    num = 5
    inputs = [num, num - 1]
    output = [x % 2 == 0 for x in inputs]
    assert_output(EvenTester(), [{"num": n} for n in inputs], output)


def test_odd():
    class OddTester(UnitTestingApp):
        @bkr.external
        def unit_test(self, num: pt.abi.Uint64, *, output: pt.abi.Bool):
            return output.set(math.odd(num.get()))

    num = 5
    inputs = [num, num - 1]
    output = [x % 2 != 0 for x in inputs]

    assert_output(OddTester(), [{"num": n} for n in inputs], output)


def test_pow10():
    class Pow10Tester(UnitTestingApp):
        @bkr.external
        def unit_test(self, num: pt.abi.Uint64, *, output: pt.abi.Uint64):
            return output.set(math.pow10(num.get()))

    num = 3
    inputs = [num]
    output = [int(10**x) for x in inputs]

    assert_output(Pow10Tester(), [{"num": n} for n in inputs], output)


def test_min():
    class MinTester(UnitTestingApp):
        @bkr.external
        def unit_test(
            self, a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ):
            return output.set(math.min(a.get(), b.get()))

    inputs = [(100, 10)]
    output = [min(a, b) for a, b in inputs]

    assert_output(MinTester(), [{"a": a, "b": b} for a, b in inputs], output)


def test_max():
    class MaxTester(UnitTestingApp):
        @bkr.external
        def unit_test(
            self, a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ):
            return output.set(math.max(a.get(), b.get()))

    inputs = [(100, 10)]
    output = [max(a, b) for a, b in inputs]

    assert_output(MaxTester(), [{"a": a, "b": b} for a, b in inputs], output)


def test_div_ceil():
    class DivCeilTester(UnitTestingApp):
        @bkr.external
        def unit_test(
            self, a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ):
            return output.set(math.div_ceil(a.get(), b.get()))

    inputs = [(100, 3)]
    output = [pymath.ceil(a / b) for a, b in inputs]

    assert_output(DivCeilTester(), [{"a": a, "b": b} for a, b in inputs], output)


def test_saturate():
    class DivCeilTester(UnitTestingApp):
        @bkr.external
        def unit_test(
            self,
            a: pt.abi.Uint64,
            b: pt.abi.Uint64,
            c: pt.abi.Uint64,
            *,
            output: pt.abi.Uint64
        ):
            return output.set(math.saturate(a.get(), b.get(), c.get()))

    inputs = [(50, 100, 20), (15, 100, 20), (150, 100, 20)]
    output = [max(min(b, a), c) for a, b, c in inputs]

    assert_output(
        DivCeilTester(), [{"a": a, "b": b, "c": c} for a, b, c in inputs], output
    )


def test_wide_factorial():
    class WideFactorialTester(UnitTestingApp):
        @bkr.external
        def unit_test(self, num: pt.abi.Uint64, *, output: pt.abi.Uint64):
            return output.set(pt.Btoi(math.wide_factorial(num.encode())))

    num = 5
    inputs = [num]
    output = [pymath.factorial(num) for num in inputs]
    assert_output(WideFactorialTester(), [{"num": num} for num in inputs], output)


def test_exponential():
    class WideFactorialTester(UnitTestingApp):
        @bkr.external
        def unit_test(
            self, num: pt.abi.Uint64, iters: pt.abi.Uint64, *, output: pt.abi.Uint64
        ):
            return output.set(math.exponential(num.get(), iters.get()))

    num = 10
    iters = 30
    inputs = [(num, iters)]
    output = [int(pymath.exp(num)) for num, _ in inputs]

    assert_output(
        WideFactorialTester(),
        [{"num": num, "iters": iters} for num, iters in inputs],
        output,
        opups=15,
    )


# def test_ln():
#   num = 10
#   expr = Log(Itob(ln(Int(num))))
#   output = [logged_int(int(pymath.log(num)))]
#   assert_output(expr, output, pad_budget=15)

# def test_log2():
#   num = 17
#   expr = Log(Itob(log2(Int(num))))
#   output = [logged_int(int(pymath.log2(num)))]
#   assert_output(expr, output, pad_budget=15)

# def test_log10():
#    num = 123123123
#    expr = Log(Itob(scaled_log10(Int(num))))
#    output = [logged_int(int(pymath.log10(num)))]
#    assert_output(expr, output)

# def test_negative_power():
#    expr = Log(negative_power(Int(100), Int(3)))
#    output = [logged_int(int(math.pow(100, -3)))]
#    assert_output(expr, output)



class BareOverwriteError(Exception):
    def __init__(self, on_complete: str):
        self.on_complete = on_complete

    def __str__(self) -> str:
        return (
            f"Tried to overwrite a bare external: {self.on_complete}.\n"
            + "If you're trying to override a default method in Application"
            + ", be sure to use the same name as the method defined (`create`, `update`, `delete`)."
        )



from pyteal import (
    Approve,
    OnComplete,
    abi,
    TealType,
    TxnType,
    Seq,
    Assert,
    Len,
    InnerTxnBuilder,
    TxnField,
    Int,
    InnerTxn,
    ScratchVar,
    Bytes,
    Log,
)
import beaker as bkr
from beaker.application import get_method_signature

algod_client = bkr.sandbox.get_algod_client()


class C2CSub(bkr.Application):
    """Sub application who's only purpose is to opt into then close out of an asset"""

    rasv = bkr.ReservedApplicationStateValue(TealType.bytes, 1)
    asv = bkr.ApplicationStateValue(TealType.bytes, default=Bytes("asv"))

    racsv = bkr.ReservedAccountStateValue(TealType.bytes, 1)
    acsv = bkr.AccountStateValue(TealType.bytes, default=Bytes("acsv"))

    @bkr.create
    def create(self):
        return Seq(self.initialize_application_state(), Approve())

    @bkr.opt_in
    def opt_in(self):
        return Seq(self.initialize_account_state(), Approve())

    @bkr.external
    def opt_in_to_asset(self, asset: abi.Asset):
        return InnerTxnBuilder.Execute(
            {
                TxnField.type_enum: TxnType.AssetTransfer,
                TxnField.xfer_asset: asset.asset_id(),
                TxnField.asset_receiver: self.address,
                TxnField.fee: Int(0),
                TxnField.asset_amount: Int(0),
            }
        )

    @bkr.external
    def return_asset(self, asset: abi.Asset, addr: abi.Account):
        return InnerTxnBuilder.Execute(
            {
                TxnField.type_enum: TxnType.AssetTransfer,
                TxnField.xfer_asset: asset.asset_id(),
                TxnField.asset_receiver: addr.address(),
                TxnField.fee: Int(0),
                TxnField.asset_amount: Int(0),
                TxnField.asset_close_to: addr.address(),
            }
        )


class C2CMain(bkr.Application):
    """Main application that handles creation of the sub app and asset and calls it"""

    # Create sub app to be precompiled before allowing TEAL generation
    sub_app: bkr.AppPrecompile = bkr.AppPrecompile(C2CSub())

    @bkr.external
    def create_sub(self, *, output: abi.Uint64):
        return Seq(
            InnerTxnBuilder.Execute(self.sub_app.get_create_config()),
            # return the app id of the newly created app
            output.set(InnerTxn.created_application_id()),
            # Try to read the global state
            sv := C2CSub.asv.get_external(output.get()),
            Log(sv.value()),
            # Opt in to the new app for funsies
            InnerTxnBuilder.Execute(
                {
                    TxnField.type_enum: TxnType.ApplicationCall,
                    TxnField.application_id: output.get(),
                    TxnField.on_completion: OnComplete.OptIn,
                }
            ),
            # Try to read the local state
            sv := C2CSub.acsv[self.address].get_external(output.get()),
            Log(sv.value()),
        )

    @bkr.internal(TealType.uint64)
    def create_asset(self, name):
        return Seq(
            InnerTxnBuilder.Execute(
                {
                    TxnField.type_enum: TxnType.AssetConfig,
                    TxnField.config_asset_name: name,
                    TxnField.config_asset_total: Int(10),
                    TxnField.config_asset_manager: self.address,
                }
            ),
            # return the newly created asset id
            InnerTxn.created_asset_id(),
        )

    @bkr.internal(TealType.none)
    def trigger_opt_in_and_xfer(self, app_id, app_addr, asset_id):
        # Call the sub app to make it opt in and xfer it some tokens
        return Seq(
            InnerTxnBuilder.Begin(),
            InnerTxnBuilder.MethodCall(
                app_id=app_id,
                method_signature=get_method_signature(C2CSub.opt_in_to_asset),
                args=[asset_id],
            ),
            InnerTxnBuilder.Next(),
            InnerTxnBuilder.SetFields(
                {
                    TxnField.type_enum: TxnType.AssetTransfer,
                    TxnField.xfer_asset: asset_id,
                    TxnField.asset_amount: Int(1),
                    TxnField.asset_receiver: app_addr,
                }
            ),
            InnerTxnBuilder.Submit(),
        )

    @bkr.internal(TealType.none)
    def trigger_return(self, app_id, asset_id):
        # Create the group txn to ask sub app to opt in and send sub app 1 token
        # Tell the sub app to send me back the stuff i sent it
        return Seq(
            InnerTxnBuilder.Begin(),
            InnerTxnBuilder.MethodCall(
                app_id=app_id,
                method_signature=get_method_signature(C2CSub.return_asset),
                args=[asset_id, self.address],
            ),
            InnerTxnBuilder.Submit(),
        )

    @bkr.external
    def create_asset_and_send(
        self, name: abi.String, sub_app_ref: abi.Application, *, output: abi.Uint64
    ):
        return Seq(
            Assert(Len(name.get())),
            # Create the asset
            (asset_id := ScratchVar()).store(self.create_asset(name.get())),
            # Get the sub app addr
            (sub_app_addr := sub_app_ref.params().address()),
            # Ask sub app to opt in, and send asset in the same group
            self.trigger_opt_in_and_xfer(
                sub_app_ref.application_id(), sub_app_addr.value(), asset_id.load()
            ),
            # Get the asset back
            self.trigger_return(sub_app_ref.application_id(), asset_id.load()),
            # Return the asset id
            output.set(asset_id.load()),
        )

    @bkr.external
    def delete_asset(self, asset: abi.Asset):
        return InnerTxnBuilder.Execute(
            {
                TxnField.type_enum: TxnType.AssetConfig,
                TxnField.config_asset: asset.asset_id(),
            }
        )


def demo():

    accts = bkr.sandbox.get_accounts()
    acct = accts.pop()
    algod_client = bkr.sandbox.get_algod_client()

    # Create main app and fund it
    app_client = bkr.client.ApplicationClient(
        algod_client, C2CMain(), signer=acct.signer
    )
    main_app_id, _, _ = app_client.create()

    print(f"Created main app: {main_app_id}")
    app_client.fund(1 * bkr.consts.algo)

    # Call the main app to create the sub app
    result = app_client.call(C2CMain.create_sub)
    print(result.tx_info)
    sub_app_id = result.return_value
    print(f"Created sub app: {sub_app_id}")

    # Call main app method to:
    #   create the asset
    #   call the sub app optin method
    #   send asset to sub app
    #   call the sub app return asset method
    sp = app_client.client.suggested_params()
    sp.flat_fee = True
    sp.fee = 1 * bkr.consts.algo
    result = app_client.call(
        C2CMain.create_asset_and_send,
        name="dope asset",
        sub_app_ref=sub_app_id,
        suggested_params=sp,
    )
    created_asset = result.return_value
    print(f"Created asset id: {created_asset}")

    result = app_client.call(
        C2CMain.delete_asset,
        asset=created_asset,
    )
    print(f"Deleted asset in tx: {result.tx_id}")


if __name__ == "__main__":
    demo()
